//****************************************************************************************************************
//                                   FastLED by Davide Rassu (aka Blame73)                                      //
//****************************************************************************************************************

// RGB (WS2812B) LED strips connected to this Arduino Uno/Nano (via I2C Mode) will show your current feedbacks
// position. This device will act as a listener (slave board) with I2C events raised by the Arduino master (the one
// with the SMC3 code)
// Wirings:
// Master UNO board -> THIS Arduino Nano: 
//      SDA -> A4, SCL -> A5, GND -> GND 
// ****************************************************************************************************************
// Changelog
// Nov 2023
// -- 1.0 Possibility to use RGB (WS2812B) LED strips
// ****************************************************************************************************************
// v.1.0 by Blame73
const unsigned int currentVersion = 100;
#include <FastLED.h>
#include <Wire.h>

// User defined settings
// ****************************************************************************************************************
#define DATA_LEDPIN1 12                         //LED output pin1 (left feedback/motor)
#define DATA_LEDPIN2 11                         //LED output pin2 (right feedback/motor)
const int I2CMaster = 5;                        //ID for the master board, MUST match the one in Ard master code
                                                    //if you change it here, modify it there accordingly
const int I2CSlave = 8;                         //ID for this slave board, MUST match the one in Ard master code
                                                    //if you change it here, modify it there accordingly
const int NUMBER_OF_STRIPS = 2;                 //Number of LED strips [1 or 2]              
const int NUMBER_OF_LEDS = 72;                  //The number of LEDs available in your LED strips
const unsigned int DEFAULT_BRIGHTNESS = 10;     //Default brightness, 255 max (keep it low to reduce current draw)
const bool REVERSE_STRIP1_DIR = true;           //Reverse LEDs direction for the first strip
const bool REVERSE_STRIP2_DIR = false;          //Reverse LEDs direction for the second strip

int COLOR_MODE = 1;                             //Select one of the following color modes:
                                                    //1 => gradient range from center, green to red
                                                    //2 => four moving dots, green to red
                                                    //3 => gradient range from center, various colors
                                                    //4 => four moving dots, various colors
int NUMBER_OF_DOTS = 6;                         //EVEN numbers! When choosing a dots mode, 
                                                    //how many dots identify the current position
// ****************************************************************************************************************

uint32_t wireTimeout = 75000U;
CRGB fl_strip1[NUMBER_OF_LEDS];
CRGB fl_strip2[NUMBER_OF_LEDS];   
CRGB fl_color_off = CRGB::Black;                //OFF 
CRGB fl_color_1 = CRGB::DarkGreen;                
CRGB fl_color_2 = CRGB::DarkRed;                
CRGB fl_color_3 = CRGB::Blue;                 
CRGB fl_color_4 = CRGB::Magenta; 
CRGB fl_colors[] = {fl_color_1, fl_color_2, fl_color_3, fl_color_4};

bool updating = false;
int colorModes = 4;
float rangePerSingleLed1;
float rangePerSingleLed2;
bool isMasterAvailable = false;
unsigned long checkMasterTimeout = 2000;        //Interval between each check when master board is not available
unsigned long startMillisCheckMaster, currentMillisCheckMaster;                
bool stripsDisabled = false;
bool enableRandomEffects = false;
int availableEffects = 5;
int currentEffect = availableEffects;

// From Arduino master, SMC3 settings. Will be set on setup()
int scaledCurrPosition1 = 512;
int scaledCurrPosition2 = 512;
int InputClipMax1 = 823;        
int InputClipMax2 = 823;
int InputClipMin1 = 200;
int InputClipMin2 = 200;

void setup()
{     
    //calculate the ranges available for the strips, excluding the clipping zones
    //once inside a clipping zone, the whole strip will be lit red
    rangePerSingleLed1 = float(NUMBER_OF_LEDS) /float(1023 - (InputClipMin1 *2));      
    rangePerSingleLed2 = float(NUMBER_OF_LEDS) /float(1023 - (InputClipMin2 *2));
    NUMBER_OF_DOTS = int(NUMBER_OF_DOTS/2);

    //FastLED setup
    FastLED.addLeds<WS2812B, DATA_LEDPIN1, GRB>(fl_strip1, NUMBER_OF_LEDS);
    if (NUMBER_OF_STRIPS > 1) 
        FastLED.addLeds<WS2812B, DATA_LEDPIN2, GRB>(fl_strip2, NUMBER_OF_LEDS);
    FastLED.setBrightness(DEFAULT_BRIGHTNESS);
    FlashStrips();

    startI2C();
    // isMasterAvailable = checkArduinoMaster();
}

void startI2C(){
    Wire.begin(I2CSlave); 
    Wire.setWireTimeout(wireTimeout, true);     // to avoid endtransmission freeze
    Wire.onReceive(receiveEvent);               // register event (calls from master)
    Serial.begin(500000);
    Serial.println("I'm alive!");  
    //Wire.requestFrom(I2CMaster, 1);           // request one byte from master board to say slave's alive
                                                // and to raise passing for parameters
}

void loop() {
    if (enableRandomEffects & !updating){
        switch (currentEffect)        
        {
        case 1:
            Effect__AddingWaves();
            break;    
        case 4:
            Effect__PhaseBeat();
            break;
        case 3:
            Effect__RainbowBeat();
            break;    
        case 2:
            Effect__BlurPhaseBeat();
            break;            
        case 5:
            Effect__MovingDot();
            break;                 
        default:
            Effect__AddingWaves();
            break;
        }
    }
}

bool checkArduinoMaster(){
    Wire.beginTransmission(I2CMaster);
    Wire.write(0);
    byte error = Wire.endTransmission();
    if (error) {
        Serial.println("Master board's ureachable");
        return false;
    }
    else{
        Serial.println("Master is alive!");
        return true;
    }        
}

// Event triggered by Arduino master board
void receiveEvent(int howMany)
{   
    if (updating)
        return;

    char param;
    int value;
    while(1 < Wire.available()) {param = Wire.read();}
    value = Wire.read();
    
    if ((param == 'L') && !stripsDisabled){                                                 // Left
        scaledCurrPosition1 = map(value, 0, 255, 0, 1023);
        if (COLOR_MODE == 1 || COLOR_MODE == 3){
            setStrip1_Range();       
        }
        else {
            setStrip1_Dots();
        } 
    }    
    else if (((param == 'R') && (NUMBER_OF_STRIPS > 1)) && !stripsDisabled) {               // Right
        scaledCurrPosition2 = map(value, 0, 255, 0, 1023);
        if (COLOR_MODE == 1 || COLOR_MODE == 3){
            setStrip2_Range();             
        }
        else {
            setStrip2_Dots();
        } 
    }
    //
    else if ((param == 'G') || (param == 'H')) {                                            // Clipping parameters
        if (param =='G'){
            InputClipMin1 = value;
            InputClipMax1 = 1023 - InputClipMin1;
            rangePerSingleLed1 = float(NUMBER_OF_LEDS) /float(1023 - (InputClipMin1 *2));             
        }
        if (param == 'H'){
            InputClipMin2 = value;
            InputClipMax2 = 1023 - InputClipMin2;
            rangePerSingleLed2 = float(NUMBER_OF_LEDS) /float(1023 - (InputClipMin2 *2));
        }
    }
    else if (param == 'P'){                                                                 // Next color mode
        updating = true;
        enableRandomEffects = false;
        COLOR_MODE = nextColorMode();
        delay(5);
        updating = false;
    }
    else if (param == 'O'){                                                                 // On/Off/Effects
        updating = true;      
        stripsDisabled = value == 0;  
        if (value == 2){
            ClearStrip1();
            if (NUMBER_OF_LEDS > 1)
                ClearStrip2();            
            currentEffect = nextEffect();
            enableRandomEffects = true;
        }
        else{
            enableRandomEffects = false;
            if (stripsDisabled){
                ClearStrip1();
                if (NUMBER_OF_LEDS > 1)
                    ClearStrip2();
            }
        }
        delay(5);
        updating = false;
    }
    else if (param == 'X'){                                                                 // Reset
        reset();
    }
    //debug
    // Serial.print(param);
    // Serial.println(value);
}

int nextColorMode() {
    return COLOR_MODE = COLOR_MODE == colorModes ? 1 : COLOR_MODE +1;
}

void SetStrip1FullColor(CRGB color){
    fill_solid(fl_strip1, NUMBER_OF_LEDS, color);
    FastLED.show();
}

void SetStrip2FullColor(CRGB color){
    fill_solid(fl_strip2, NUMBER_OF_LEDS, color);
    FastLED.show();
}

void SetStrip1GradientColors(){
    fill_gradient_RGB(fl_strip1, NUMBER_OF_LEDS, fl_color_2, fl_color_1, fl_color_3, fl_color_4);
    blur1d(fl_strip1, NUMBER_OF_LEDS, 64);
    FastLED.show();
}

void SetStrip2GradientColors(){    
    fill_gradient_RGB(fl_strip2, NUMBER_OF_LEDS, fl_color_2, fl_color_1, fl_color_3, fl_color_4);
    blur1d(fl_strip2, NUMBER_OF_LEDS, 64);
    FastLED.show();
}

void ClearStrip1(){
    SetStrip1FullColor(fl_color_off);
    FastLED.show();
}

void ClearStrip2(){
    SetStrip2FullColor(fl_color_off);
    FastLED.show();
}

void setStrip1_Dots(){
    //when inside clipping zone draw all strip red
    if ((scaledCurrPosition1 > InputClipMax1) || (scaledCurrPosition1 < InputClipMin1)){
        SetStrip1FullColor(fl_color_2);
    }
    else {
        //calculate the current LED based on its position inside the valid range zone (excluding clipping zone) 
        int currentLed = int((scaledCurrPosition1 - InputClipMin1) * rangePerSingleLed1);
        if (REVERSE_STRIP1_DIR) {
            currentLed = abs(currentLed - NUMBER_OF_LEDS);
        }
        //set a colored background
        if (COLOR_MODE < 3){
            fill_gradient_RGB(fl_strip1, NUMBER_OF_LEDS, fl_color_2, fl_color_1, fl_color_2);
        }
        else{
            fill_gradient_RGB(fl_strip1, NUMBER_OF_LEDS, fl_color_2, fl_color_1, fl_color_3, fl_color_4);;
        }
        //and turn off what's outside the current LED (with some range around it)
        for (int i=0; i < currentLed - NUMBER_OF_DOTS; i++) {
            fl_strip1[i] = fl_color_off;
        }
        for (int i=currentLed + NUMBER_OF_DOTS; i < NUMBER_OF_LEDS; i++) {
            fl_strip1[i] = fl_color_off;
        }
        blur1d(fl_strip1, NUMBER_OF_LEDS, 64);
        FastLED.show();
    }
}

void setStrip2_Dots(){
    if ((scaledCurrPosition2 > InputClipMax2) || (scaledCurrPosition2 < InputClipMin2)){
        SetStrip2FullColor(fl_color_2);
    }
    else {
        int currentLed = int((scaledCurrPosition2 - InputClipMin2) * rangePerSingleLed2);
        if (REVERSE_STRIP2_DIR) {
            currentLed = abs(currentLed - NUMBER_OF_LEDS);
        }

        if (COLOR_MODE < 3){
            fill_gradient_RGB(fl_strip2, NUMBER_OF_LEDS, fl_color_2, fl_color_1, fl_color_2);
        }
        else{
            fill_gradient_RGB(fl_strip2, NUMBER_OF_LEDS, fl_color_2, fl_color_1, fl_color_3, fl_color_4);;
        }

        for (int i=0; i < currentLed -NUMBER_OF_DOTS; i++) {
            fl_strip2[i] = fl_color_off;
        }
        for (int i=currentLed +NUMBER_OF_DOTS; i < NUMBER_OF_LEDS; i++) {
            fl_strip2[i] = fl_color_off;
        }
        blur1d(fl_strip2, NUMBER_OF_LEDS, 64);
        FastLED.show();
    }
}

void setStrip1_Range(){
//same logic as for Dot functions but turning off the opposite half
    if ((scaledCurrPosition1 > InputClipMax1) || (scaledCurrPosition1 < InputClipMin1)){
        SetStrip1FullColor(fl_color_2);
    }
    else {
        int currentLed = int((scaledCurrPosition1 - InputClipMin1) * rangePerSingleLed1);
        if (REVERSE_STRIP1_DIR) {
            currentLed = abs(currentLed - NUMBER_OF_LEDS);
        }

        if (COLOR_MODE == 1){
            fill_gradient_RGB(fl_strip1, NUMBER_OF_LEDS, fl_color_2, fl_color_1, fl_color_2);
        }
        else{
            fill_gradient_RGB(fl_strip1, NUMBER_OF_LEDS, fl_color_2, fl_color_1, fl_color_3, fl_color_4);
        }        

        if (currentLed > NUMBER_OF_LEDS /2){
            for (int i=currentLed +NUMBER_OF_DOTS; i < NUMBER_OF_LEDS; i++) {
                fl_strip1[i] = fl_color_off;
            } 
            for (int i=0; i < NUMBER_OF_LEDS /2 -NUMBER_OF_DOTS; i++) {
                fl_strip1[i] = fl_color_off;
            } 
        }
        else if (currentLed < NUMBER_OF_LEDS /2){
            for (int i=0; i < currentLed -NUMBER_OF_DOTS; i++) {
                fl_strip1[i] = fl_color_off;
            }        
            for (int i=NUMBER_OF_LEDS /2 +NUMBER_OF_DOTS; i < NUMBER_OF_LEDS; i++) {
                fl_strip1[i] = fl_color_off;
            } 
        }
        else {  //center position, draw it as dots
            setStrip1_Dots();   
        }
        blur1d(fl_strip1, NUMBER_OF_LEDS, 64);
        FastLED.show();
    }
}

void setStrip2_Range(){
    if ((scaledCurrPosition2 > InputClipMax2) || (scaledCurrPosition2 < InputClipMin2)){
        SetStrip2FullColor(fl_color_2);
    }
    else {
        int currentLed = int((scaledCurrPosition2 - InputClipMin2) * rangePerSingleLed2);
        if (REVERSE_STRIP2_DIR) {
            currentLed = abs(currentLed - NUMBER_OF_LEDS);
        }

        if (COLOR_MODE == 1){
            fill_gradient_RGB(fl_strip2, NUMBER_OF_LEDS, fl_color_2, fl_color_1, fl_color_2);
        }
        else{
            fill_gradient_RGB(fl_strip2, NUMBER_OF_LEDS, fl_color_2, fl_color_1, fl_color_3, fl_color_4);
        }

        if (currentLed > NUMBER_OF_LEDS /2){
            for (int i=currentLed +NUMBER_OF_DOTS; i < NUMBER_OF_LEDS; i++) {
                fl_strip2[i] = fl_color_off;
            } 
            for (int i=0; i < NUMBER_OF_LEDS /2 -NUMBER_OF_DOTS; i++) {
                fl_strip2[i] = fl_color_off;
            } 
        }
        else if (currentLed < NUMBER_OF_LEDS /2){
            for (int i=0; i < currentLed -NUMBER_OF_DOTS; i++) {
                fl_strip2[i] = fl_color_off;
            }        
            for (int i=NUMBER_OF_LEDS /2 +NUMBER_OF_DOTS; i < NUMBER_OF_LEDS; i++) {
                fl_strip2[i] = fl_color_off;
            } 
        }
        else {
            setStrip2_Dots();
        }
        blur1d(fl_strip2, NUMBER_OF_LEDS, 64);
        FastLED.show();
    }
}

void FlashStrips(){
    SetStrip1GradientColors();    
    if (NUMBER_OF_STRIPS > 1) {SetStrip2GradientColors();}
    delay(333);
    ClearStrip1();
    if (NUMBER_OF_STRIPS > 1) {ClearStrip2();}
    delay(333);
    SetStrip1GradientColors();    
    if (NUMBER_OF_STRIPS > 1) {SetStrip2GradientColors();}
    delay(333);
    ClearStrip1();
    if (NUMBER_OF_STRIPS > 1) {ClearStrip2();}
}

void reset(){ 
    asm volatile ("jmp 0");
}

int nextEffect(){
    return currentEffect == availableEffects ? 1 : currentEffect +1;
    // if (currentEffect == availableEffects)
    //     return 1;
    // else 
    //     return currentEffect +1;
}


//****************************************************************************************************************
//                                                 Custom Effects                                               //
//****************************************************************************************************************

void Effect__AddingWaves(){
    uint8_t posBeat  = beatsin8(38, 0, NUMBER_OF_LEDS - 1, 0, 0);
    uint8_t posBeat2 = beatsin8(60, 0, NUMBER_OF_LEDS - 1, 0, 0);
    uint8_t posBeat3 = beatsin16(38, 0, NUMBER_OF_LEDS - 1, 0, 127);
    uint8_t posBeat4 = beatsin16(60, 0, NUMBER_OF_LEDS - 1, 0, 127);

    uint8_t colBeat  = beatsin8(45, 0, 255, 0, 0);
    fl_strip1[(posBeat + posBeat2) / 2]  = CHSV(colBeat, 255, 255);
    fl_strip2[(posBeat + posBeat2) / 2]  = CHSV(colBeat, 255, 255);
    fl_strip2[(posBeat3 + posBeat4) / 2]  = CHSV(colBeat, 255, 255);
    fl_strip1[(posBeat3 + posBeat4) / 2]  = CHSV(colBeat, 255, 255);

    fadeToBlackBy(fl_strip1, NUMBER_OF_LEDS, 10);
    fadeToBlackBy(fl_strip2, NUMBER_OF_LEDS, 10);
    FastLED.show();
}

void Effect__PhaseBeat(){
    uint8_t sinBeat   = beatsin8(30, 0, NUMBER_OF_LEDS - 1, 0, 0);
    uint8_t sinBeat2  = beatsin8(30, 0, NUMBER_OF_LEDS - 1, 0, 85);
    uint8_t sinBeat3  = beatsin8(30, 0, NUMBER_OF_LEDS - 1, 0, 170);

    fl_strip1[sinBeat]   = CRGB::Green;
    fl_strip1[sinBeat2]  = CRGB::Red;
    fl_strip1[sinBeat3]  = CRGB::White;
    fl_strip2[sinBeat]   = CRGB::Red;
    fl_strip2[sinBeat2]  = CRGB::Blue;
    fl_strip2[sinBeat3]  = CRGB::Red;

    fadeToBlackBy(fl_strip1, NUMBER_OF_LEDS, 10);
    fadeToBlackBy(fl_strip2, NUMBER_OF_LEDS, 10);
    FastLED.show();
}

void Effect__RainbowBeat(){
    uint16_t beatA = beatsin16(50, 0, 255);
    uint16_t beatB = beatsin16(22, 0, 255);

    fill_rainbow(fl_strip1, NUMBER_OF_LEDS, (beatA+beatB)/2, 8);
    fill_rainbow(fl_strip2, NUMBER_OF_LEDS, (beatA+beatB)/2, 8);    
    FastLED.show();
}

void Effect__BlurPhaseBeat(){
    uint8_t sinBeat1  = beatsin8(30, 0, NUMBER_OF_LEDS - 1, 0, 0);
    uint8_t sinBeat2  = beatsin8(30, 0, NUMBER_OF_LEDS - 1, 0, 60);
    uint8_t sinBeat3  = beatsin8(30, 0, NUMBER_OF_LEDS - 1, 0, 120);
    uint8_t sinBeat4  = beatsin8(30, 0, NUMBER_OF_LEDS - 1, 0, 170);

    fl_strip1[sinBeat1]  = CRGB::Green;
    fl_strip1[sinBeat2]  = CRGB::DarkCyan;
    fl_strip1[sinBeat3]  = CRGB::Red;
    fl_strip1[sinBeat4]  = CRGB::Yellow;
    fl_strip2[sinBeat1]  = CRGB::Green;
    fl_strip2[sinBeat2]  = CRGB::DarkCyan;
    fl_strip2[sinBeat3]  = CRGB::Red;
    fl_strip2[sinBeat4]  = CRGB::Yellow;

    EVERY_N_MILLISECONDS(1){
        for(int i = 0; i < 4; i++) {
            blur1d(fl_strip1, NUMBER_OF_LEDS, 128);
            blur1d(fl_strip2, NUMBER_OF_LEDS, 50);
        }
    }
    fadeToBlackBy(fl_strip1, NUMBER_OF_LEDS, 10);
    fadeToBlackBy(fl_strip2, NUMBER_OF_LEDS, 10);
    FastLED.show();
}

void Effect__MovingDot(){
    uint16_t sinBeat = beatsin16(30, 0, NUMBER_OF_LEDS - 1, 0, 0);

    fl_strip1[sinBeat] = fl_color_1;
    fl_strip2[sinBeat] = fl_color_1;
    fadeToBlackBy(fl_strip1, NUMBER_OF_LEDS, 10);
    fadeToBlackBy(fl_strip2, NUMBER_OF_LEDS, 10);

    FastLED.show();
}