//****************************************************************************************************************
// SMC3.ini is a basic Motor PID driver designed for motion simulators with up to 3 motors (written for UNO R3)
//****************************************************************************************************************
// --Modified by Davide Rassu (aka Blame73) for MODE2 Cytron MD30C (may work with Cytron MDDS30 too)
// May 2020
// --Modified forces and brakes that were inverted on half side of the range. Revised Sept 2023
// Sept 2023
// --Modified clipping zone in order to reduce speed with and without rev force being set
//     Note: should you want to reverse motors direction (instead of slowering them down) when inside clipping zone
//     comment and uncomment where needed (search for MOD by Blame73)
// --Solved a problem where, once off-limits, motors couldn't turn back on
// Sept 2023
// --Added linear scaler that works for Cytrons too. Revised Oct 2023
// Oct 2023
// --Added possibility to use a TM1638 module to monitor ranges and eventually the scaler
//      available pages:
    //  1 > Current distances from nearest clipping zones
    //  2 > Current feedback values        
    //  3 > Max values reached
    //  4 > Min values reached
    //  5 > Upper and lower clipping zones reached counters
    //  6 > Min distance from upper clipping zone reached (negative values when inside clipping zone)
    //  7 > Min distance from lower clipping zone reached (negative values when inside clipping zone)
    //  8 > Change LED strips color mode
    //  9 > Reset values
    // 21 > (button 2 + button 1) > Store value red from scaler (the one on the right side) as minimum value
    // 32 > (button 3 + button 2) > Reset min scaler value to default (defaultScalerLowerLimit)
    // 89 > (button 8 + button 9) > Enable/Disable LED strips
// Nov 2023
// --Added possibility to use RGB (WS2812B) LED strips
//****************************************************************************************************************
// v.1.06 by Blame73
const unsigned int currentVersion = 106;
//****************************************************************************************************************
// THIS SCRIPT HAS BEEN MODIFIED TO BE A MODE2 SKETCH SOLELY WITH NO MOTOR3 SUPPORT
//****************************************************************************************************************


//****************************************************************************************************************
//                                               Scaler by Blame73                                              //
//****************************************************************************************************************
// Uncomment the following line to enable scaling via two modes for motion values:
// MODE1, via potentiometer using analogue input AN3
// MODE2, digitally by TM1638 buttons (two buttons to increase/decrease scaling values)
// #define ENABLE_SCALING

#ifdef ENABLE_SCALING
    // User defined settings
    // ***********************************************************************************************************
    const int defaultScalerLowerLimit = 250;    // Not to scale too much. It's the lower side in the 0-1024 
                                                // pot scaler range; the lower you set this number the higher
                                                // it will scale. don't go below 25! 250 is a 50% scaler
    bool scaleByPot = true;                    
    // ***********************************************************************************************************

    float scaledValue;
    bool showingScaler = false;  
    unsigned long scalerTimeout = 8000;
    unsigned long startMillisScaler, currentMillisScaler; 
    int scaler = defaultScalerLowerLimit;
    int scalerLowerLimit, previousScalerValue; 
#endif

const int adjustedFeedback1 = 0;
const int adjustedFeedback2 = 0;

//****************************************************************************************************************
//                                               FastLED by Blame73                                             //
//****************************************************************************************************************
// You can attach the strips to this Arduino or send data to another Arduino UNO/Nano (via I2C Mode) 
// Uncomment the following line to enable the use of an RGB (WS2812B) LED strip 
// #define ENABLE_FASTLED

#ifdef ENABLE_FASTLED
    #include <Wire.h>
    // In case you enabled FastLed, choose one of the two modes:
    // MODE 1, *REMOVED to have less overload on this board* - the strips were attached to this Arduino board
    // MODE 2, the strips are attached to another Arduino (Uno or Nano -TESTED- or whatever works with you)
    // Wirings for MODE2:
    // This UNO -> Arduino Nano: 
    //      SDA -> A4, SCL -> A5, GND -> GND     
    // This UNO -> another Arduino UNO: 
    //      SDA -> SDA, SCL -> SCL, GND -> GND 
    const int I2CMaster = 5;                        //I2C ID for this board, master 
    const int I2CSlave = 8;                         //I2C ID for the other board, slave
    uint32_t wireTimeout = 100000U;                   
    unsigned long fastledTimeout = 50;              //Interval between each command sent to the slave, once per strip (to lower slave overload)
    unsigned long startMillisFastled, currentMillisFastled;
    bool stripsDisabled = false;
    bool stripsDisabledbyUser = false;
    bool isSlaveAvailable = false;
    bool sendLeftValue = true;
    byte retryCounter = 0;
#endif


//****************************************************************************************************************
//                                               TM1638 by Blame73                                              //
//****************************************************************************************************************
// Uncomment the following line to enable the use of a TM1638 module to monitor your rig working ranges
// #define ENABLE_TM1638

#ifdef ENABLE_TM1638
    #include <TM1638.h>    
    // TM1638 -> Arduino UNO connection: 
    // VCC -> 5v, GND -> GND 
    // DIO, CLK, STB according to the following three lines:
    #define DIO 6
    #define CLK 7
    #define STB 8
    TM1638 moduleTM1638(DIO, CLK, STB);

    // User defined settings
    // ***********************************************************************************************************
    const unsigned int moduleBrightness = 3;        // Display brightness, from 0 (min brightness) to 7
    const unsigned long btnPressedInterval = 150;   // Timer for button pressed check
    const int defaultPage = 1;                      // TM1638 starting and default page
    bool autoTurnOffDisplay = false;                // Turns the display off after some time if ranges don't change
                                                    // Re-enabled by changed position or button pressed
    #define inverted_colors                         // Uncomment if your red and green leds colors are inverted    
    const bool showAlertMessages = false;           // Show messages when too close to cutoff limits or beyond them    
    const int maxCutoffWarningDistance = 50;        // How close to upper cutoff limits to get a warning message
    const int minCutoffWarningDistance = 50;        // How close to lower cutoff limits to get a warning message
    // ***********************************************************************************************************

    // TM1638 LEDs
    #ifdef inverted_colors
        #define RED_LED TM1638_COLOR_GREEN
        #define GREEN_LED TM1638_COLOR_RED
    #else
        #define RED_LED TM1638_COLOR_RED
        #define GREEN_LED TM1638_COLOR_GREEN
    #endif
    #define ORANGE_LED GREEN_LED + RED_LED
    #define BLANK_LED 0
    
    // TM1638 pages
    unsigned int TM1638_PAGE = defaultPage;
    unsigned int TM1638_PREV_PAGE = defaultPage;      
    byte buttons;   
    #define button1 0b00000001    // 1
    #define button2 0b00000010    // 2
    #define button3 0b00000100    // 4
    #define button4 0b00001000    // 8
    #define button5 0b00010000    // 16
    #define button6 0b00100000    // 32
    #define button7 0b01000000    // 64
    #define button8 0b10000000    // 128 
    #define button1_8 0b10000001  // 129
    #define button1_2 0b00000011  // 3
    #define button2_3 0b00000110  // 6
    #define button6_7 0b01100000  // 96
    #define button8_9 0b11000000  // 192
    
    int scaledCurrPosition1 = 512;
    int scaledCurrPosition2 = 512;
    int previousScaledCurrPosition1 = 512;
    int previousScaledCurrPosition2 = 512;
    int maxPot1, minPot1, maxPot2, minPot2;
    long unsigned startMillis, currentMillis;
    int rangePerLed_WorkingZone, rangePerLed_ClippingZone;
    int maxClipLimitReachedPot1Count, minClipLimitReachedPot1Count, maxClipLimitReachedPot2Count, minClipLimitReachedPot2Count;
    bool stopCountingMaxClipLimit1 = false;
    bool stopCountingMinClipLimit1 = false;
    bool stopCountingMaxClipLimit2 = false;
    bool stopCountingMinClipLimit2 = false;
    int limitDistancePot1, limitDistancePot2;
    int maxLimitDistancePot1, maxLimitDistancePot2, minLimitDistancePot1, minLimitDistancePot2;  
    
    // Timers
    bool showingMessage = false;    // Using a timer to show messages because delay() would freeze the app
    unsigned long messageTimeout = 2000;
    unsigned long startMillisMessage, currentMillisMessage;
    unsigned long displayHandlerTimeout = 500;
    unsigned long startMillisDisplayHandler, currentMillisDisplayHandler;

    // Running Average (to disable module if not used); one value added each second
    #include "RunningAverage.h"
    RunningAverage range_RA(120);   // two minutes range
    bool isDisplayOff = false;

    // Messages
    const char splashMsg[]  = "SMC3 MON";
    const char warningMsg[] = "WARNING";
    const char errorMsg[]   = "ERROR";
#endif


// Since I'm using those pins for the TM1638 module I have removed motor3 (also because I wasn't using it and this lightens calculations)

// Uncomment the following line to reverse the direction of Motor 1.
// #define REVERSE_MOTOR1

//    COMMAND SET:
//
//    []              		Drive all motors to defined stop positions and hold there
//    [Axx],[Bxx],[Cxx]   Send position updates for Motor 1,2,3 where xx is the binary position limitted to range 0-1024
//    [Dxx],[Exx],[Fxx]		Send the Kp parameter for motor 1,2,3 where xx is the Kp value multiplied by 100
//    [Gxx],[Hxx],[Ixx]		Send the Ki parameter for motor 1,2,3 where xx is the Ki value multiplied by 100
//    [Jxx],[Kxx],[Lxx]		Send the Kd parameter for motor 1,2,3 where xx is the Kd value multiplied by 100
//    [Mxx],[Nxx],[Oxx]		Send the Ks parameter for motor 1,2,3 where xx is the Ks value
//    [Pxy],[Qxy],[Rxy]		Send the PWMmin and PWMmax values x is the PWMmin and y is PWMmax each being in range 0-255
//    [Sxy],[Txy],[Uxy]   Send the Motor Min/Max Limits (x) and Input Min/Max Limits (y) (Note same value used for Min and Max)         
//    [Vxy],[Wxy],[Xxy]   Send the Feedback dead zone (x) and the PWM reverse duty (y) for each motor
//    [rdx]          		  Read a value from the controller where x is the code for the parameter to read    
//    [ena]			          Enable all motors
//    [sav]			          Save parameters to non-volatile memory
//    [ver]               Send the SMC3 software version
//

//    Arduino UNO Pinouts Used
//    ------------------------
//    A0 - Motor 1 Feedback
//    A1 - Motor 2 Feedback
//    A5 - Motion scaler Feedback
//    2 - Motor 1 H-Bridge ENA 
//    3 - Motor 1 H-Bridge ENB
//    4 - Motor 2 H-Bridge ENA
//    5 - Motor 2 H-Bridge ENB
//    6 - TM1638 DIO
//    7 - TM1638 CLK
//    8 - TM1638 STB
//    9 - Motor 1 PWM       
//    10 - Motor 2 PWM 
//    SDA - Arduino UNO/NANO Slave
//    SCL - Arduino UNO/NANO Slave


// BOF preprocessor bug prevention - leave me at the top of the arduino-code
#if 1
__asm volatile ("nop");
#endif

#include <EEPROM.h>
#include <SoftwareSerial.h>

// defines for setting and clearing register bits
#ifndef cbi
    #define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
    #define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#define COM0 0                          // hardware Serial Port
#define COM1 1                          // Software Serial Port
#define START_BYTE '['                  // Start Byte for serial commands
#define END_BYTE ']'                    // End Byte for serial commands
#define PROCESS_PERIOD_uS 250           // 244 -> 4096 per second approx
#define TIMER_UPPER_LIMIT 4294966000    // Used to check for timer wraparound 
#define TIMER_LOWER_LIMIT 1000          // Used to check for timer wraparound


unsigned long TimesUp=0;                // Counter used to see if it's time to calculate next PID update

int Feedback1 = 512;
int Feedback2 = 512;
int Feedback3 = 512;
int Target1 = 512;
int Target2 = 512;
int Target3 = 512;
int PotInput = 512;

unsigned int RxByte[2]={0};         // Current byte received from each of the two comm ports
int BufferEnd[2]={-1};              // Rx Buffer end index for each of the two comm ports
unsigned int RxBuffer[5][2]={0};    // 5 byte Rx Command Buffer for each of the two comm ports
unsigned long LoopCount	= 0;        // unsigned 32bit, 0 to 4,294,967,295 to count times round loop
unsigned long LastCount = 0;        // loop counter the last time it was read so can calc delta
int errorcount	= 0;                // serial receive error detected by invalid packet start/end bytes
unsigned int CommsTimeout = 0;      // used to reduce motor power if there has been no comms for a while
int PowerScale = 7;                // used to divide(shift) PID result changes when in low power

// 
//
//                    +----+              +------+
//                  +-|PWR |--------------| USB  |-----+
//                  | |    |              |      |     |
//                  | +----+              +------+    o|SCL    Arduino UNO/NANO Slave FastLED
//                  |                                 o|SDA    Arduino UNO/NANO Slave FastLED
//                  |                                 o|AREF
//                  |                                 o|GND   
//                NC|o                                o|13     
//             IOREF|o   +----+                       o|12     
//             RESET|o   |    |                       o|11    
//              3.3V|o   |    |                       o|10~    PWM Motor 2
//                5V|o   |    |    Arduino            o|9~     PWM Motor 1
//               GND|o   |    |      UNO              o|8***   TM1638 STB       
//               GND|o   |    |                        |
//               VIN|o   |    |                       o|7***   TM1638 CLK       
//                  |    |    |                       o|6***   TM1638 DIO       
//  Motor 1 Pot   A0|o   |    |                       o|5~     Motor 2 ENB
//  Motor 2 Pot   A1|o   |    |                       o|4      Motor 2 ENA
//                A2|o   |    |                       o|3~     Motor 1 ENB
//                A3|o   |    |                       o|2      Motor 1 ENA
//                A4|o   |    |                       o|1
//  Motion Scaler A5|o   +-/\-+                       o|0
//                  +-\                      /---------+
//                     \--------------------/
//
// 

int OutputPort =PORTD;          // read the current port D bit mask
const int ENApin1 = 2;          // ENA output pin for Motor H-Bridge 1 (ie PortD bit position) 
const int ENBpin1 = 3;          // ENB output pin for Motor H-Bridge 1 (ie PortD bit position)
const int ENApin2 = 4;          // ENA output pin for Motor H-Bridge 2 (ie PortD bit position)
const int ENBpin2 = 5;          // ENB output pin for Motor H-Bridge 2 (ie PortD bit position)

const int PWMpin1 = 9;          // PWM output pin for Motor 1   
const int PWMpin2 = 10;         // PWM output pin for Motor 2    
const uint8_t FeedbackPin1 = A0;    // Motor 1 feedback pin
const uint8_t FeedbackPin2 = A1;    // Motor 2 feedback pin
const uint8_t PotScalerPin = A5;    // User adjustable POT used to scale the motion (if enabled)

int DeadZone1 = 0;              // feedback deadzone		
int DeadZone2 = 0;              // feedback deadzone
int DeadZone3 = 0;              // feedback deadzone

int CenterOffset1 = 0;          // Adjust center offset of feedback position
int CenterOffset2 = 0;          // Adjust center offset of feedback position
int CenterOffset3 = 0;          // Adjust center offset of feedback position

int CutoffLimitMax1 = 1000;     // The position beyond which the motors are disabled
int CutoffLimitMax2 = 1000;     // The position beyond which the motors are disabled
int CutoffLimitMax3 = 1000;     // The position beyond which the motors are disabled
int CutoffLimitMin1 = 23;       // The position beyond which the motors are disabled
int CutoffLimitMin2 = 23;       // The position beyond which the motors are disabled
int CutoffLimitMin3 = 23;       // The position beyond which the motors are disabled

int InputClipMax1 = 923;        // The input position beyond which the target input is clipped
int InputClipMax2 = 923;        // The input position beyond which the target input is clipped
int InputClipMax3 = 923;        // The input position beyond which the target input is clipped
int InputClipMin1 = 100;        // The input position beyond which the target input is clipped
int InputClipMin2 = 100;        // The input position beyond which the target input is clipped
int InputClipMin3 = 100;        // The input position beyond which the target input is clipped

int LiftFactor1 = 0;            // was 10 - Increase PWM when driving motor in direction it has to work harder 
int LiftFactor2 = 0;            // was 10 - Increase PWM when driving motor in direction it has to work harder

int PIDProcessDivider = 1;      // divider for the PID process timer
int PIDProcessCounter = 0;
int SerialFeedbackCounter = 0;
int SerialFeedbackEnabled = 0;
int SerialFeedbackPort = 0;

int Ks1 = 1;
long Kp1_x100 = 100;		    //initial value
long Ki1_x100 = 40;
long Kd1_x100 = 40;
int Ks2 = 1;
long Kp2_x100 = 420;
long Ki2_x100 = 40;
long Kd2_x100 = 40;
int Ks3 = 1;
int Kp3_x100 = 420;
int Ki3_x100 = 40;
int Kd3_x100 = 40;
int PWMout1 = 0;
int PWMout2 = 0;
int PWMout3 = 0;

int Disable1 = 1;               //Motor stop flag
int Disable2 = 1;               //Motor stop flag
int Disable3 = 1;               //Motor stop flag
int PWMoffset1 = 50;
int PWMoffset2 = 50;
int PWMoffset3 = 50;
int PWMmax1 = 100;
int PWMmax2 = 100;
int PWMmax3 = 100;
int PWMrev1 = 200;
int PWMrev2 = 200;
int PWMrev3 = 200;
unsigned int Timer1FreqkHz = 25;   // PWM freq used for Motor 1 and 2
unsigned int Timer2FreqkHz = 31;   // PWM freq used for Motor 3


//****************************************************************************************************************
//    Setup the PWM pin frequencies
//          
//****************************************************************************************************************


void setPwmFrequency(int pin, int divisor) 
{
    int mode;
    if(pin == 5 || pin == 6 || pin == 9 || pin == 10) 
    {
        switch(divisor) 
        {
            case 1: mode = 0x01; break;
            case 8: mode = 0x02; break;
            case 64: mode = 0x03; break;
            case 256: mode = 0x04; break;
            case 1024: mode = 0x05; break;
            default: return;
        }
        if(pin == 5 || pin == 6) 
        {
            TCCR0B = (TCCR0B & 0b11111000) | mode;
        } 
        else 
        {
            TCCR1B = (TCCR1B & 0b11111000) | mode;
        }
    } 
    else 
    {
        if(pin == 3 || pin == 11) 
        {
            switch(divisor) 
            {
                case 1: mode = 0x01; break;
                case 8: mode = 0x02; break;
                case 32: mode = 0x03; break;
                case 64: mode = 0x04; break;
                case 128: mode = 0x05; break;
                case 256: mode = 0x06; break;
                case 1024: mode = 0x7; break;
                default: return;
            }
            TCCR2B = (TCCR2B & 0b11111000) | mode;
        }
    }
}

void InitialisePWMTimer1(unsigned int Freq)     // Used for pins 9 and 10
{
	uint8_t wgm = 8;    //setting the waveform generation mode to 8 - PWM Phase and Freq Correct
	TCCR1A = (TCCR1A & B11111100) | (wgm & B00000011);
	TCCR1B = (TCCR1B & B11100111) | ((wgm & B00001100) << 1);
	TCCR1B = (TCCR1B & B11111000) | 0x01;    // Set the prescaler to minimum (ie divide by 1)

	unsigned int CountTOP;

        CountTOP = (F_CPU / 2) / Freq;    // F_CPU is the oscillator frequency - Freq is the wanted PWM freq

        // Examples of CountTOP:
	    //  400 = 20000Hz  -> 8MHz / Freq = CountTOP
        //  320 = 25000Hz
        //  266 = 30075Hz 

	ICR1 = CountTOP;          // Set the TOP of the count for the PWM
}


void InitialisePWMTimer2(unsigned int Freq) 
{
    int Timer2DivideMode;
    
    if (Freq>=15000)   
    {
       Timer2DivideMode=0x01;   // Anything above 15000 set divide by 1 which gives 31250Hz PWM freq
    }
    else
    {
       Timer2DivideMode=0x02;   // Anything below 15000 set divide by 8 (mode 2) which gives 3906Hz PWM freq
    }
    TCCR2B = (TCCR2B & 0b11111000) | Timer2DivideMode;
}


void MyPWMWrite(uint8_t pin, uint8_t val)
{
    #define OCR1A_MEM      0x88
    #define OCR1B_MEM      0x8A

    pinMode(pin, OUTPUT);
	
    //casting "val" to be larger so that the final value (which is the partially
    //the result of multiplying two potentially high value int16s) will not truncate
    uint32_t tmp = val;
	
    if (val == 0)
        digitalWrite(pin, LOW);
    else if (val == 255)
        digitalWrite(pin, HIGH);
    else
    {
        uint16_t regLoc16 = 0;
        uint16_t top;
		
        switch(pin)
        {
            case 9:
                sbi(TCCR1A, COM1A1);
                regLoc16 = OCR1A_MEM;
                top = ICR1;                  //Timer1_GetTop();
                break;
            case 10:
                sbi(TCCR1A, COM1B1);
                regLoc16 = OCR1B_MEM;
                top = ICR1;                  //Timer1_GetTop();
                break;
        }
        tmp=(tmp*top)/255;
        _SFR_MEM16(regLoc16) = tmp;       //(tmp*top)/255;
    }		
}


void DisableMotor1();
void DisableMotor2();

//****************************************************************************************************************
//    Arduino setup subroutine called at startup/reset
//          
//****************************************************************************************************************


/// @brief 
void setup()
{
    //Read all stored Parameters from EEPROM

    ReadEEProm();

    Serial.begin(500000);     //Default 500000  //115200
    // set the data rate for the SoftwareSerial port
    OutputPort=PORTD;
    
    pinMode(ENApin1, OUTPUT);
    pinMode(ENBpin1, OUTPUT);
    pinMode(PWMpin1, OUTPUT);    
    MyPWMWrite(PWMpin1,0);      //analogWrite(PWMpin1, 0);
    DisableMotor1();
        
    pinMode(ENApin2, OUTPUT);
    pinMode(ENBpin2, OUTPUT);
    pinMode(PWMpin2, OUTPUT);
    MyPWMWrite(PWMpin2,0);      //analogWrite(PWMpin2, 0); 
    DisableMotor2();  

    // Note that the base frequency for pins 3, 9, 10, and 11 is 31250 Hz
    //setPwmFrequency(PWMpin1, 1);     // Divider can be 1, 8, 64, 256
    //setPwmFrequency(PWMpin2, 1);     // Divider can be 1, 8, 64, 256
    //setPwmFrequency(PWMpin3, 1);     // Divider can be 1, 8, 64, 256
    InitialisePWMTimer1(Timer1FreqkHz * 1000);        // PWM Freq for Motor 1 and 2 can be set quite precisely - pass the desired freq and nearest selected
    InitialisePWMTimer2(Timer2FreqkHz * 1000);        // Motor 3 PWM Divider can only be 1, 8, 64, 256 giving Freqs 31250,3906, 488 and 122 Hz

    //CLKPR = 0x80;   // These two lines halve the processor clock
    //CLKPR = 0x01;

    // set analogue prescale to 16
    sbi(ADCSRA,ADPS2);
    cbi(ADCSRA,ADPS1);
    cbi(ADCSRA,ADPS0);
    
    //****************************************************************************************************************
    //                                               TM1638 by Blame73                                              //
    //****************************************************************************************************************
    #ifdef ENABLE_TM1638    
        moduleTM1638.setupDisplay(true, moduleBrightness);     
        showTM1638Message(splashMsg, 2000);
        resetRangeValues();
    #endif

    //****************************************************************************************************************
    //                                               FastLED by Blame73                                             //
    //****************************************************************************************************************
    #ifdef ENABLE_FASTLED        
        Wire.begin(I2CMaster);
        Wire.setWireTimeout(wireTimeout, true);
        Wire.onReceive(receiveEvent);                       //Just in case, but unused
        isSlaveAvailable = checkSlaveBoardStatus();                 
    #endif
}


void WriteEEPRomWord(int address, int intvalue)
{
    int low,high;
    high=intvalue/256;
    low=intvalue-(256*high);
    EEPROM.write(address,high);
    EEPROM.write(address+1,low);
}

int ReadEEPRomWord(int address)
{
    int low,high, returnvalue;
    high=EEPROM.read(address);
    low=EEPROM.read(address+1);
    returnvalue=(high*256)+low;
    return returnvalue;
}

//****************************************************************************************************************
//    Write all configurable parameters to EEPROM
//          
//****************************************************************************************************************


void WriteEEProm()
{
    EEPROM.write(0,114);
    EEPROM.write(1,CutoffLimitMin1);
    EEPROM.write(2,InputClipMin1);    
    EEPROM.write(5,DeadZone1);
    EEPROM.write(6,CutoffLimitMin2);
    EEPROM.write(7,InputClipMin2);
    EEPROM.write(10,DeadZone2);
    WriteEEPRomWord(11,Kp1_x100);    // **** Even though these variables are long the actual values stored are only 0 to 1000 so OK ****
    WriteEEPRomWord(13,Ki1_x100);
    WriteEEPRomWord(15,Kd1_x100);
    WriteEEPRomWord(17,Kp2_x100);
    WriteEEPRomWord(19,Ki2_x100);
    WriteEEPRomWord(21,Kd2_x100);
    EEPROM.write(23,PWMoffset1);
    EEPROM.write(24,PWMoffset2);
    EEPROM.write(25,PWMmax1);
    EEPROM.write(26,PWMmax2);
    EEPROM.write(27,PWMrev1);
    EEPROM.write(28,PWMrev2);        
    EEPROM.write(29,PWMrev3);
    EEPROM.write(31,CutoffLimitMin3);
    EEPROM.write(32,InputClipMin3);
    EEPROM.write(35,DeadZone3);
    WriteEEPRomWord(36,Kp3_x100);
    WriteEEPRomWord(38,Ki3_x100);
    WriteEEPRomWord(40,Kd3_x100);
    EEPROM.write(42,PWMoffset3);
    EEPROM.write(43,PWMmax3);
    WriteEEPRomWord(44,Ks1);
    WriteEEPRomWord(46,Ks2);    
    WriteEEPRomWord(48,Ks3);
    EEPROM.write(50,constrain(PIDProcessDivider,1,10));
    EEPROM.write(51,Timer1FreqkHz);
    EEPROM.write(52,Timer2FreqkHz);   
}

//****************************************************************************************************************
//    Read all configurable parameters from the EEPROM.  
//          
//****************************************************************************************************************


void ReadEEProm()
{
    int evalue = EEPROM.read(0);
    if(evalue != 114) //EEProm was not set before, set default values
    {
        WriteEEProm();
        return;
    }
    CutoffLimitMin1 = EEPROM.read(1);
    InputClipMin1 = EEPROM.read(2);
    CutoffLimitMax1 = 1023-CutoffLimitMin1;
    InputClipMax1 = 1023-InputClipMin1;
    
    DeadZone1=EEPROM.read(5);
    CutoffLimitMin2 = EEPROM.read(6);
    InputClipMin2 = EEPROM.read(7);
    CutoffLimitMax2 = 1023-CutoffLimitMin2;
    InputClipMax2 = 1023-InputClipMin2;

    DeadZone2=EEPROM.read(10);
    Kp1_x100=ReadEEPRomWord(11);
    Ki1_x100=ReadEEPRomWord(13);
    Kd1_x100=ReadEEPRomWord(15);
    Kp2_x100=ReadEEPRomWord(17);
    Ki2_x100=ReadEEPRomWord(19);
    Kd2_x100=ReadEEPRomWord(21);
    PWMoffset1=EEPROM.read(23);
    PWMoffset2=EEPROM.read(24);
    PWMmax1=EEPROM.read(25);
    PWMmax2=EEPROM.read(26);
    PWMrev1=EEPROM.read(27);
    PWMrev2=EEPROM.read(28);
    
    PWMrev3=EEPROM.read(29);
    CutoffLimitMin3 = EEPROM.read(31);
    InputClipMin3 = EEPROM.read(32);
    CutoffLimitMax3 = 1023-CutoffLimitMin3;
    InputClipMax3 = 1023-InputClipMin3;
    DeadZone3=EEPROM.read(35);
    Kp3_x100=ReadEEPRomWord(36);
    Ki3_x100=ReadEEPRomWord(38);
    Kd3_x100=ReadEEPRomWord(40);
    PWMoffset3=EEPROM.read(42);
    PWMmax3=EEPROM.read(43);
    
    Ks1=ReadEEPRomWord(44);
    Ks2=ReadEEPRomWord(46);
    Ks3=ReadEEPRomWord(48);

    PIDProcessDivider=EEPROM.read(50);
    PIDProcessDivider=constrain(PIDProcessDivider,1,10);
    Timer1FreqkHz=EEPROM.read(51);
    Timer2FreqkHz=EEPROM.read(52);
}



//****************************************************************************************************************
//    Send two bytes via serial in response to a request for information
//          
//****************************************************************************************************************

void SendTwoValues(int id, int v1, int v2, int ComPort)
{
    if (ComPort==0)
    {
        Serial.write(START_BYTE);
        Serial.write(id);
        Serial.write(v1);
        Serial.write(v2);
        Serial.write(END_BYTE);
    }
}


//****************************************************************************************************************
//    Send a single word value via serial in response to a request for information
//          
//****************************************************************************************************************

void SendValue(int id, int value, int ComPort)
{
    int low,high;

    high=value/256;
    low=value-(high*256);

    if (ComPort==0)
    {
        Serial.write(START_BYTE);
        Serial.write(id);
        Serial.write(high);
        Serial.write(low);
        Serial.write(END_BYTE);
    }
}


//****************************************************************************************************************
//    Calculate how many PID calculations have been performed since last requested
//          
//****************************************************************************************************************

int DeltaLoopCount()
{
    unsigned long Delta;
    
    if ( (LastCount==0) || ((LoopCount-LastCount)>32000) )
    {
        Delta = 0;
        LastCount = LoopCount;
    }
    else
    {
         Delta = LoopCount-LastCount;
         LastCount  = LoopCount; 
    }    
    return (int)Delta;
}


//****************************************************************************************************************
//    Process the incoming serial commands from the specified comm port
//          
//****************************************************************************************************************


void ParseCommand(int ComPort)
{
    CommsTimeout = 0;    // reset the comms timeout counter to indicate we are getting packets
    
    switch (RxBuffer[0][ComPort]) 
    {
        case 'A':
            Target1=(RxBuffer[1][ComPort]*256)+RxBuffer[2][ComPort];
#ifdef REVERSE_MOTOR1
            Target1=1023-Target1;
#endif
            //************************************************************************* 
            //                           Scaler by Blame73                           //
            //*************************************************************************
            // a different linear scaler that uses a previously calculated Scaled Value
            // by Blame73
            #ifdef ENABLE_SCALING                        
                Target1 = int(float(Target1 -512) /scaledValue) +512;
            #endif
            //************************************************************************* 
            Target1 += adjustedFeedback1;

            if (Target1>InputClipMax1) { Target1=InputClipMax1; }
            else if (Target1<InputClipMin1) { Target1=InputClipMin1; }
            break;
        case 'B':
            Target2=(RxBuffer[1][ComPort]*256)+RxBuffer[2][ComPort];


            //************************************************************************* 
            //                           Scaler by Blame73                           //
            //*************************************************************************
            // a different linear scaler that uses a previously calculated Scaled Value
            // by Blame73
            #ifdef ENABLE_SCALING 
                Target2 = int(float(Target2 -512) /scaledValue) +512;
            #endif 
            //************************************************************************* 
            Target2 += adjustedFeedback2;

            if (Target2>InputClipMax2) { Target2=InputClipMax2; }
            else if (Target2<InputClipMin2) { Target2=InputClipMin2; }
            break;
        case 'C':
            Target3=(RxBuffer[1][ComPort]*256)+RxBuffer[2][ComPort];
            if (Target3>InputClipMax3) { Target3=InputClipMax3; }
            else if (Target3<InputClipMin3) { Target3=InputClipMin3; }
            break;
        case 'r':
            if (RxBuffer[1][ComPort]=='d')      // rd - Read a value from the ArduinoPID - next byte identifies value to read
            {
                switch (RxBuffer[2][ComPort]) 
                {
                    case 'A':
#ifdef REVERSE_MOTOR1
                        SendTwoValues('A',(1023-Feedback1)/4,(1023-Target1)/4,ComPort);
#else
                        SendTwoValues('A',Feedback1/4,Target1/4,ComPort);
#endif
                        break;
                    case 'B':
                        SendTwoValues('B',Feedback2/4,Target2/4,ComPort);
                        break;
                    case 'C':
                        SendTwoValues('C',Feedback3/4,Target3/4,ComPort);
                        break;
                    case 'a':
                        SendTwoValues('a',PIDProcessDivider*16 + Disable1+(Disable2*2)+(Disable3*4),constrain(PWMout1,0,255),ComPort);
                        break;
                    case 'b':
                        SendTwoValues('b',PIDProcessDivider*16 + Disable1+(Disable2*2)+(Disable3*4),constrain(PWMout2,0,255),ComPort);
                        break;
                    case 'c':
                        SendTwoValues('c',PIDProcessDivider*16 + Disable1+(Disable2*2)+(Disable3*4),constrain(PWMout3,0,255),ComPort);
                        break;
                    case 'D':
                        SendValue('D',Kp1_x100,ComPort);
                        break;
                    case 'E':
                        SendValue('E',Kp2_x100,ComPort);
                        break;
                    case 'F':
                        SendValue('F',Kp3_x100,ComPort);
                        break;
                    case 'G':
                        SendValue('G',Ki1_x100,ComPort);
                        break;
                    case 'H':
                        SendValue('H',Ki2_x100,ComPort);
                        break;
                    case 'I':
                        SendValue('I',Ki3_x100,ComPort);
                        break;
                    case 'J':
                        SendValue('J',Kd1_x100,ComPort);
                        break;
                    case 'K':
                        SendValue('K',Kd2_x100,ComPort);
                        break;
                    case 'L':
                        SendValue('L',Kd3_x100,ComPort);
                        break;
                    case 'M':
                        SendValue('M',Ks1,ComPort);
                        break;
                    case 'N':
                        SendValue('N',Ks2,ComPort);
                        break;
                    case 'O':
                        SendValue('O',Ks3,ComPort);
                        break;
                    case 'P':
                        SendTwoValues('P',PWMoffset1,PWMmax1,ComPort);
                        break;
                    case 'Q':
                        SendTwoValues('Q',PWMoffset2,PWMmax2,ComPort);
                        break;
                    case 'R':
                        SendTwoValues('R',PWMoffset3,PWMmax3,ComPort);
                        break;
                    case 'S':
                        SendTwoValues('S',CutoffLimitMin1,InputClipMin1,ComPort);
                        break;
                    case 'T':
                        SendTwoValues('T',CutoffLimitMin2,InputClipMin2,ComPort);
                        break;
                    case 'U':
                        SendTwoValues('U',CutoffLimitMin3,InputClipMin3,ComPort);
                        break;
                    case 'V':
                        SendTwoValues('V',DeadZone1,PWMrev1,ComPort);
                        break;
                    case 'W':
                        SendTwoValues('W',DeadZone2,PWMrev2,ComPort);
                        break;
                    case 'X':
                        SendTwoValues('X',DeadZone3,PWMrev3,ComPort);
                        break;
                    case 'Y':
                        SendTwoValues('Y',PIDProcessDivider*16 + Disable1+(Disable2*2)+(Disable3*4),0,ComPort);  // Second byte not yet used
                        break;
                    case 'Z':
                        SendValue('Z',DeltaLoopCount(),ComPort);
//                        SendValue('Z',errorcount,ComPort);            // *** TEMP CODE FOR TESTING
                        break;
                    case '~':
                        SendTwoValues('~',Timer1FreqkHz,Timer2FreqkHz,ComPort);  // PWM Frequencies to set
                        break;
                     case '?':
                        break;
//                    default:
                }
            }
            break;
        case 'D':
            Kp1_x100=(RxBuffer[1][ComPort]*256)+RxBuffer[2][ComPort];
            break;
        case 'E':
            Kp2_x100=(RxBuffer[1][ComPort]*256)+RxBuffer[2][ComPort];
            break;
        case 'F':
            Kp3_x100=(RxBuffer[1][ComPort]*256)+RxBuffer[2][ComPort];
            break;
        case 'G':
            Ki1_x100=(RxBuffer[1][ComPort]*256)+RxBuffer[2][ComPort];
            break;
        case 'H':
            Ki2_x100=(RxBuffer[1][ComPort]*256)+RxBuffer[2][ComPort];
            break;
        case 'I':
            Ki3_x100=(RxBuffer[1][ComPort]*256)+RxBuffer[2][ComPort];
            break;
        case 'J':
            Kd1_x100=(RxBuffer[1][ComPort]*256)+RxBuffer[2][ComPort];
            break;
        case 'K':
            Kd2_x100=(RxBuffer[1][ComPort]*256)+RxBuffer[2][ComPort];
            break;
        case 'L':
            Kd3_x100=(RxBuffer[1][ComPort]*256)+RxBuffer[2][ComPort];
            break;
        case 'M':
            Ks1=(RxBuffer[1][ComPort]*256)+RxBuffer[2][ComPort];       
            break;
        case 'N':
            Ks2=(RxBuffer[1][ComPort]*256)+RxBuffer[2][ComPort];
            break;
        case 'O':
            Ks3=(RxBuffer[1][ComPort]*256)+RxBuffer[2][ComPort];
            break;
        case 'P':
            PWMoffset1=RxBuffer[1][ComPort];
            PWMmax1=RxBuffer[2][ComPort];
            break;
        case 'Q':
            PWMoffset2=RxBuffer[1][ComPort];
            PWMmax2=RxBuffer[2][ComPort];
            break;
        case 'R':
            PWMoffset3=RxBuffer[1][ComPort];
            PWMmax3=RxBuffer[2][ComPort];
            break;
        case 'S':
            CutoffLimitMin1=RxBuffer[1][ComPort];
            CutoffLimitMax1=1023-CutoffLimitMin1;
            InputClipMin1=RxBuffer[2][ComPort];
            InputClipMax1=1023-InputClipMin1;
            break;
        case 'T':
            CutoffLimitMin2=RxBuffer[1][ComPort];
            CutoffLimitMax2=1023-CutoffLimitMin2;
            InputClipMin2=RxBuffer[2][ComPort];
            InputClipMax2=1023-InputClipMin2;
            break;
        case 'U':
            CutoffLimitMin3=RxBuffer[1][ComPort];
            CutoffLimitMax3=1023-CutoffLimitMin3;
            InputClipMin3=RxBuffer[2][ComPort];
            InputClipMax3=1023-InputClipMin3;
            break;
        case 'V':
            DeadZone1=RxBuffer[1][ComPort];
            PWMrev1=RxBuffer[2][ComPort];
            break;
        case 'W':
            DeadZone2=RxBuffer[1][ComPort];
            PWMrev2=RxBuffer[2][ComPort];
            break;
        case 'X':
            DeadZone3=RxBuffer[1][ComPort];
            PWMrev3=RxBuffer[2][ComPort];
            break;
        case 'Z':
            PIDProcessDivider=constrain(RxBuffer[1][ComPort],1,10);
            // ???=RxBuffer[2][ComPort];   // second byte not yet used
            break;
        case '~':
            Timer1FreqkHz=RxBuffer[1][ComPort];
            Timer2FreqkHz=RxBuffer[2][ComPort];   
            InitialisePWMTimer1(Timer1FreqkHz * 1000);
            InitialisePWMTimer2(Timer2FreqkHz * 1000);
            break;
        case 'm':
            if (RxBuffer[1][ComPort]=='o' && RxBuffer[2][ComPort]=='1')   // Monitor motor 1 (auto sends position and feedback data to host
            {
                SerialFeedbackEnabled= (ComPort << 4) | 1;   // Motor 1
            }
            if (RxBuffer[1][ComPort]=='o' && RxBuffer[2][ComPort]=='2')   // Monitor motor 2 (auto sends position and feedback data to host
            {
                SerialFeedbackEnabled= (ComPort << 4) | 2;   // Motor 2
            }
            if (RxBuffer[1][ComPort]=='o' && RxBuffer[2][ComPort]=='3')   // Monitor motor 3 (auto sends position and feedback data to host
            {
                SerialFeedbackEnabled= (ComPort << 4) | 3;   // Motor 3
            }
            if (RxBuffer[1][ComPort]=='o' && RxBuffer[2][ComPort]=='0')   // Switch off auto monitoring data feedback
            {
                SerialFeedbackEnabled=0;
            }
            break;
        case 's':
            if (RxBuffer[1][ComPort]=='a' && RxBuffer[2][ComPort]=='v')   // Save all settings to EEPROM
            {
                WriteEEProm();
            }
            break;
        case 'v':
            if (RxBuffer[1][ComPort]=='e' && RxBuffer[2][ComPort]=='r')   // Send back the SMC3 software version
            {
                SendValue('v', currentVersion, ComPort);     // Software Version - divide by 100 to get version - ie 101= ver1.01
            }
            break;
        case 'e':
            if (RxBuffer[1][ComPort]=='n' && RxBuffer[2][ComPort]=='a')   // Enable all motors
            {
                Disable1=0;
                Disable2=0;
                Disable3=0;
                OutputPort &= ~(1 << ENBpin1);        // Unset Motor1 In 2
                OutputPort &= ~(1 << ENBpin2);        // Unset Motor2 In 2
                PORTD = OutputPort;
                OutputPort |= 1 << ENBpin1;           // Set Motor1 In 2
                OutputPort |= 1 << ENBpin2;           // Set Motor2 In 2
                PORTD = OutputPort;
            }
            else if (RxBuffer[1][ComPort]=='n' && RxBuffer[2][ComPort]=='1')   // Enable motor 1
            {
                Disable1=0;
                OutputPort &= ~(1 << ENBpin1);        // Unset Motor1 In 2
                PORTD = OutputPort;
                OutputPort |= 1 << ENBpin1;           // Set Motor1 In 2
                PORTD = OutputPort;
            }
            else if (RxBuffer[1][ComPort]=='n' && RxBuffer[2][ComPort]=='2')   // Enable motor 2
            {
                Disable2=0;
                OutputPort &= ~(1 << ENBpin2);        // Unset Motor2 In 2
                PORTD = OutputPort;
                OutputPort |= 1 << ENBpin2;           // Set Motor2 In 2
                PORTD = OutputPort;
            }
            break;
        case '?':
            break;
//        default: 
    }  
  

}

//****************************************************************************************************************
//    Check the incoming serial data to see if a command has been received
//          
//****************************************************************************************************************


void CheckSerial0()
{
    while(Serial.available()) 
    {
        if(BufferEnd[COM0]==-1)
        {
            RxByte[COM0] = Serial.read();
            if(RxByte[COM0] != START_BYTE){BufferEnd[COM0]=-1;errorcount++;}else{BufferEnd[COM0]=0;}
        }
        else
        {
            RxByte[COM0] = Serial.read();
            RxBuffer[BufferEnd[COM0]][COM0]=RxByte[COM0];
            BufferEnd[COM0]++;
            if(BufferEnd[COM0] > 3)
            {
                if(RxBuffer[3][COM0]==END_BYTE)
                {
                    ParseCommand(COM0);
                }
                else
                {
                    errorcount++;
                }
                BufferEnd[COM0]=-1;
            }
        }
    }
}


//****************************************************************************************************************
//    Set the Motor output pins to drive the motor in required direction and speed
//          
//****************************************************************************************************************


void SetOutputsMotor1()
{
    // Cilp Zone MAX
    if((Feedback1 > InputClipMax1) && (PWMrev1 != 0))  //                                                                     ------------- MOD By Blame73 for Cytron MD30C
    {             
      // lower the speed when inside clip input zone
      // if you want to reverse motors direction when inside clip input zone 
       // replace the following three lines with the commented ones       
       PWMout1 = 255 - (PWMout1 / 3);
       OutputPort &= 1 << ENApin1;    // Set Motor1 In 1
       MyPWMWrite(PWMpin1, PWMout1); 
       
       //PWMout1 = PWMrev1;
       //OutputPort |= 1 << ENApin1;  // Unset Motor1 In 1
       //MyPWMWrite(PWMpin1, PWMout1); 
    }
    else if((Feedback1 > InputClipMax1) && (PWMrev1 == 0))  //                                                                ------------- MOD By Blame73 for Cytron MD30C (if rev not set, lower the speed anyway)
    {
       PWMout1 = 255 - (PWMout1 / 3);
       OutputPort &= 1 << ENApin1;    // Set Motor1 In 1
       MyPWMWrite(PWMpin1, PWMout1); 
    }
    
    // Cilp Zone MIN
    else if((Feedback1 < InputClipMin1) && (PWMrev1 != 0))  //                                                                ------------- MOD By Blame73 for Cytron MD30C
    {
       // lower the speed when inside clip input zone
       // if you want to reverse motors direction when inside clip input zone 
       // replace the following three lines with the commented ones       
       PWMout1 = (PWMout1 / 3);
       OutputPort |= 1 << ENApin1;  // Unset Motor1 In 1
       MyPWMWrite(PWMpin1, PWMout1);       
       
       //PWMout1 = PWMrev1;
       //OutputPort &= ~(1 << ENApin1);        // Unset Motor1 In 1
       //MyPWMWrite(PWMpin1, PWMout1); 
    }
    else if((Feedback1 < InputClipMin1) && (PWMrev1 == 0))  //                                                                ------------- MOD By Blame73 for Cytron MD30C (if rev not set, lower the speed anyway)
    {
       PWMout1 = (PWMout1 / 3);
       OutputPort |= 1 << ENApin1;  // Unset Motor1 In 1
       MyPWMWrite(PWMpin1, PWMout1); 
    }   
     
    else if((Target1 > (Feedback1 + DeadZone1)) || (Target1 < (Feedback1 - DeadZone1)))
    {
        if (PWMout1 >= 0)  
        {                                    
            // Drive Motor Forward 
            PWMout1+=PWMoffset1;
            if(PWMout1 > (PWMmax1+LiftFactor1)){PWMout1=PWMmax1+LiftFactor1;}
            OutputPort |= 1 << ENApin1;           // Set Motor1 In 1
            MyPWMWrite(PWMpin1, PWMout1);    // Motor driven when PWM=0 (unset)                                               ------------- MOD By Blame73 for Cytron MD30C
        }  
        else 
        {                                              
            // Drive Motor Backwards 
            PWMout1 = abs(PWMout1);
            PWMout1+=PWMoffset1;
            if(PWMout1 > PWMmax1){PWMout1=PWMmax1;}
            OutputPort &= ~(1 << ENApin1);        // Unset Motor1 In 1
            MyPWMWrite(PWMpin1, PWMout1);        // Motor driven when PWM=1 (set)
        }
    }
    else
    {
        // Brake Motor 
        //OutputPort &= ~(1 << ENApin1);        // Unset Motor1 In 1
        OutputPort |= 1 << ENApin1;
        
        PWMout1=PWMoffset1;
        MyPWMWrite(PWMpin1, 0);
    }
    PORTD = OutputPort;
}


void SetOutputsMotor2()
{
   // Cilp Zone MAX
    if((Feedback2 > InputClipMax2) && (PWMrev2 != 0))  //                                                                     ------------- MOD By Blame73 for Cytron MD30C
    {
       // lower the speed when inside clip input zone
       // if you want to reverse motors direction when inside clip input zone 
       // replace the following three lines with the commented ones    
       PWMout2 = 255 - (PWMout2 / 3);
       OutputPort &= 1 << ENApin2;    // Set Motor2 In 1
       MyPWMWrite(PWMpin2, PWMout2);       
       
       //PWMout2 = PWMrev2;
       //OutputPort |= 1 << ENApin2;  // Unset Motor2 In 1
       //MyPWMWrite(PWMpin2, PWMout2); 
    }
    else if((Feedback2 > InputClipMax2) && (PWMrev2 == 0))  //                                                                ------------- MOD By Blame73 for Cytron MD30C (if rev not set, lower the speed anyway)
    {
       PWMout2 = 255 - (PWMout2 / 3);
       OutputPort &= 1 << ENApin2;    // Set Motor2 In 1
       MyPWMWrite(PWMpin2, PWMout2); 
    }
    
    // Cilp Zone MIN
    else if((Feedback2 < InputClipMin2) && (PWMrev2 != 0))  //                                                                ------------- MOD By Blame73 for Cytron MD30C
    {
       // lower the speed when inside clip input zone
       // if you want to reverse motors direction when inside clip input zone 
       // replace the following three lines with the commented ones    
       PWMout2 = (PWMout2 / 3);
       OutputPort |= 1 << ENApin2;  // Unset Motor2 In 1
       MyPWMWrite(PWMpin2, PWMout2);       
       
       //PWMout2 = PWMrev2;
       //OutputPort &= ~(1 << ENApin2);        // Unset Motor2 In 1
       //MyPWMWrite(PWMpin2, PWMout2); 
    }
    else if((Feedback2 < InputClipMin2) && (PWMrev2 == 0))  //                                                                ------------- MOD By Blame73 for Cytron MD30C (if rev not set, lower the speed anyway)
    {
       PWMout2 = (PWMout2 / 3);
       OutputPort |= 1 << ENApin2;  // Unset Motor2 In 1
       MyPWMWrite(PWMpin2, PWMout2); 
    }   
     
    else if((Target2 > (Feedback2 + DeadZone2)) || (Target2 < (Feedback2 - DeadZone2)))
    {
        if (PWMout2 >= 0)  
        {                                    
            // Drive Motor Forward 
            PWMout2+=PWMoffset2;
            if(PWMout2 > (PWMmax2+LiftFactor2)){PWMout2=PWMmax2+LiftFactor2;}
            OutputPort |= 1 << ENApin2;           // Set Motor2 In 1
            MyPWMWrite(PWMpin2, PWMout2);    // Motor driven when PWM=0 (unset)                                               ------------- MOD By Blame73 for Cytron MD30C
        }  
        else 
        {                                              
            // Drive Motor Backwards 
            PWMout2 = abs(PWMout2);
            PWMout2+=PWMoffset2;
            if(PWMout2 > PWMmax2){PWMout2=PWMmax2;}
            OutputPort &= ~(1 << ENApin2);        // Unset Motor2 In 1
            MyPWMWrite(PWMpin2, PWMout2);        // Motor driven when PWM=1 (set)
        }
    }
    else
    {
        // Brake Motor 
        //OutputPort &= ~(1 << ENApin2);        // Unset Motor2 In 1
        OutputPort |= 1 << ENApin2;
        
        PWMout2=PWMoffset2;
        MyPWMWrite(PWMpin2, 0);
    }
    PORTD = OutputPort;
}



//****************************************************************************************************************
//    Calculate the motor PID output
//    Based on http://brettbeauregard.com/blog/2011/04/improving-the-beginners-pid-introduction/
//    Re-written to eliminate the use of float calculations to significantly increase calculation speed. 
//          
//****************************************************************************************************************



int CalcMotor1PID(int TargetPosition, int CurrentPosition)   
{
    static int Error=0;                                                     
    static long pTerm_x100=0;
    static long dTerm_x100=0;
    static long iTerm_x100=0;
    static int CumError=0;
    static int LastPosition=0;
    static int DeltaPosition=0;
    static int KdFilterCount=0;

    Error = TargetPosition - CurrentPosition;
    if (abs(Error)<=DeadZone1)
    {
        CumError = 0;
    }
    else
    {
        CumError += Error;
        CumError = constrain(CumError,-1024,1024);            
    }         

    pTerm_x100 = Kp1_x100 * (long)Error;                    // Error can only be between +/-1023 and Kp1_100 is constrained to 0-1000 so can work with type long
    iTerm_x100 = (Ki1_x100 * (long)CumError);               // was >>6

    KdFilterCount++;
    if (KdFilterCount >= Ks1)                               // Apply a level of filtering to Kd parameter to help reduce motor noise
    {
        DeltaPosition = (CurrentPosition - LastPosition);
        LastPosition = CurrentPosition;
        dTerm_x100 = (Kd1_x100 * (long)DeltaPosition);      // was <<5
        KdFilterCount=0;    
    }
    
    return constrain((pTerm_x100 + iTerm_x100 - dTerm_x100) >> PowerScale ,-255,255);   //  the /128 (PowerScale) is an approximation to bring x100 terms back to units.  Accurate/consistent enough for this function.
}

int CalcMotor2PID(int TargetPosition, int CurrentPosition)   
{
    static int Error=0;                                                     
    static long pTerm_x100=0;
    static long dTerm_x100=0;
    static long iTerm_x100=0;
    static int CumError=0;
    static int LastPosition=0;
    static int DeltaPosition=0;
    static int KdFilterCount=0;

    Error = TargetPosition - CurrentPosition;
    if (abs(Error)<=DeadZone2)
    {
        CumError = 0;
    }
    else
    {
        CumError += Error;
        CumError = constrain(CumError,-1024,1024);            // Maybe Try 512 as the limits??????
    }

    DeltaPosition = (CurrentPosition - LastPosition);
    LastPosition = CurrentPosition;
             

    pTerm_x100 = Kp2_x100 * (long)Error;                    // Error can only be between +/-1023 and Kp1_100 is constrained to 0-1000 so can work with type long
    iTerm_x100 = (Ki2_x100 * (long)CumError);        

    KdFilterCount++;
    if (KdFilterCount >= Ks2)                               // Apply a level of filtering to Kd parameter to help reduce motor noise
    {
        DeltaPosition = (CurrentPosition - LastPosition);
        LastPosition = CurrentPosition;
        dTerm_x100 = (Kd2_x100 * (long)DeltaPosition);    
        KdFilterCount=0;    
    }
    
    return constrain((pTerm_x100 + iTerm_x100 - dTerm_x100) >> PowerScale ,-255,255);   //  the /128 is an approximation to bring x100 terms back to units.  Accurate/consistent enough for this function.
}

//****************************************************************************************************************
//    Disable the Motors - ie put the brakes on
//    Used when motor feedback is outside allowed range to minimise damage
//          
//****************************************************************************************************************

void DisableMotor1()
{
    Disable1 = 1;

    OutputPort &= ~(1 << ENApin1);
    OutputPort &= ~(1 << ENBpin1);

    PORTD = OutputPort;

    MyPWMWrite(PWMpin1, 0);

}

void DisableMotor2()
{
    Disable2 = 1;

    OutputPort &= ~(1 << ENApin2);
    OutputPort &= ~(1 << ENBpin2);

    PORTD = OutputPort;

    MyPWMWrite(PWMpin2, 0);

}


void TogglePin()
{
   static int PinOut=0;
   PinOut=1-PinOut;
   digitalWrite(8, PinOut); 
}


//****************************************************************************************************************
//    Main Arduino Program Loop
//          
//****************************************************************************************************************

void loop()
{

    // Enable all motors

    Disable1=0;
    Disable2=0;

    OutputPort &= ~(1 << ENBpin1);        // Unset Motor1 In 2
    OutputPort &= ~(1 << ENBpin2);        // Unset Motor2 In 2
    PORTD = OutputPort;
    OutputPort |= 1 << ENBpin1;           // Set Motor1 In 2
    OutputPort |= 1 << ENBpin2;           // Set Motor2 In 2
    PORTD = OutputPort;
     

    // Initialise the PID ready timer

    TimesUp = micros();
    PIDProcessCounter=0;
    
    // Main Program loop

    while (1==1) 
    {

        // Wait until it's time and then update PID calcs for first motor

        while ((micros() - TimesUp) < PROCESS_PERIOD_uS) { ; }
        TimesUp += PROCESS_PERIOD_uS;
        TogglePin();                      // Used for testing to monitor PID timing on Oscilloscope

        PIDProcessCounter++;

        if (PIDProcessCounter >= PIDProcessDivider)
        {
            PIDProcessCounter=0;
            
            // Check and Update Motor 1 drive
    
            Feedback1 = analogRead(FeedbackPin1);

            //********************************************************************* 
            //                         Scaler by Blame73                         //
            //*********************************************************************
            // a different scaler implementation           
            #ifdef ENABLE_SCALING                                                                    
                // Calculate the scaled value to only do it once for all
                // round to 5 to avoid some flickering
                if (scaleByPot){
                        scaler = round(analogRead(PotScalerPin) /5) *5;
                    if (scaler < scalerLowerLimit){
                        scaler = scalerLowerLimit;
                    }
                    else if (scaler > 1000){
                        scaler = 1000;
                    }
                    // when scaler changes (with some tolerance to avoid pot flickering), show the scaler page
                    if ((scaler > previousScalerValue + 30) || (scaler < previousScalerValue - 30)){
                        previousScalerValue = scaler;
                        TM1638_PREV_PAGE = TM1638_PAGE;
                        TM1638_PAGE = 76;    
                        showingScaler = true;      
                        // If a module is connected and display is off, turn it back on                   
                        #ifdef ENABLE_TM1638
                            if (autoTurnOffDisplay)
                                range_RA.addValue(50000);   // Add an high number to the average so that it will take some time to turn the display back off again
                            isDisplayOff = false;    
                        //     #ifdef ENABLE_FASTLED
                        //     if (!stripsDisabledbyUser)
                        //         setStripsOnOff(isDisplayOff);
                        //         Serial.println("1466");
                        //     #endif
                        #endif                  
                    }       
                }

                currentMillisScaler = millis();
                if ((currentMillisScaler - startMillisScaler > scalerTimeout) && showingScaler){
                    showingScaler = false;
                    TM1638_PAGE = defaultPage;
                    //TM1638_PAGE = TM1638_PREV_PAGE;
                    startMillisScaler = currentMillisScaler;
                }  

                scaledValue = sqrt(scaler * 0.001);
                Feedback1 = (Feedback1 -512) /scaledValue +512;          
            #endif         
            //*********************************************************************        
            Feedback1 += adjustedFeedback1;

            if ((Feedback1 > CutoffLimitMax1) || (Feedback1 < CutoffLimitMin1)) { DisableMotor1(); } 
            PWMout1=CalcMotor1PID(Target1,Feedback1);
            if (Disable1==0) 
            { 
                SetOutputsMotor1(); 
            }
            else
            {
                PWMout1=0;
            }

            // Check and Update Motor 2 drive

            Feedback2 = analogRead(FeedbackPin2);

            //********************************************************************* 
            //                         Scaler by Blame73                         //
            //*********************************************************************
            #ifdef ENABLE_SCALING
                  Feedback2 = (Feedback2 -512) /scaledValue +512;
            #endif
            //*********************************************************************        
            Feedback2 += adjustedFeedback2;

            if ((Feedback2 > CutoffLimitMax2) || (Feedback2 < CutoffLimitMin2)) { DisableMotor2(); } 
            PWMout2=CalcMotor2PID(Target2,Feedback2);
            if (Disable2==0) 
            { 
                SetOutputsMotor2();
            }
            else
            {
                PWMout2=0;
            }           


            //********************************************************************* 
            //                         TM1638 by Blame73                         //
            //*********************************************************************         
            #ifdef ENABLE_TM1638            
                checkRangesAndButtons();
                // Handle the display every n millis
                currentMillisDisplayHandler = millis();
                if ((currentMillisDisplayHandler - startMillisDisplayHandler > displayHandlerTimeout) && autoTurnOffDisplay){
                    displayHandler();
                    startMillisDisplayHandler = currentMillisDisplayHandler;
                } 
                // if ((millis() % 1000 == 0) && autoTurnOffDisplay) {
                //     displayHandler();
                // }
            #endif
            //*********************************************************************

            
            //********************************************************************* 
            //                         FastLED by Blame73                        //
            //*********************************************************************
            #ifdef ENABLE_FASTLED
                if (!stripsDisabled && !stripsDisabledbyUser){
                    currentMillisFastled = millis();
                    if (currentMillisFastled - startMillisFastled > fastledTimeout){
                        // Send values over I2C 
                        if (isSlaveAvailable){
                            // sending L&R alternate values not to overload the slave board
                            // values are mapped to bytes in the 0/255 range
                            if (sendLeftValue)
                                sendValuesToSlave('L', map(Feedback1, 0, 1023, 0, 255));
                            else 
                                sendValuesToSlave('R', map(Feedback2, 0, 1023, 0, 255));
                        }
                        else {
                            // check if it's back alive every n loops
                            if (retryCounter == 50){
                                isSlaveAvailable = checkSlaveBoardStatus();                                    
                                retryCounter = 0;
                            }
                            retryCounter++;
                        }
                        startMillisFastled = currentMillisFastled;
                        sendLeftValue = !sendLeftValue;
                    } 
                }
            #endif
            //*********************************************************************

            LoopCount++;
        }
        
        // Check if there is any received serial data and process (Hardware UART)

        CheckSerial0();

        SerialFeedbackCounter++;
        if (SerialFeedbackCounter >= 80)  // every 20ms send back position, pwm and motor status updates if enabled 
        {
            SerialFeedbackPort = (SerialFeedbackEnabled >> 4) & 0x01;
            if  ((SerialFeedbackEnabled & 0x03) == 1)     // Monitor Motor 1
            {
            #ifdef REVERSE_MOTOR1
                           SendTwoValues('A',(1023-Feedback1)/4,(1023-Target1)/4,SerialFeedbackPort);
            #else
                           SendTwoValues('A',Feedback1/4,Target1/4,SerialFeedbackPort);
            #endif
               SendTwoValues('a',PIDProcessDivider*16 + Disable1+(Disable2*2)+(Disable3*4),constrain(PWMout1,0,255),SerialFeedbackPort);
            }
            else if  ((SerialFeedbackEnabled & 0x03) == 2)     // Monitor Motor 2
            {
               SendTwoValues('B',Feedback2/4,Target2/4,SerialFeedbackPort);
               SendTwoValues('b',PIDProcessDivider*16 + Disable1+(Disable2*2)+(Disable3*4),constrain(PWMout2,0,255),SerialFeedbackPort);
            }
            else if  ((SerialFeedbackEnabled & 0x03) == 3)     // Monitor Motor 3
            {
               SendTwoValues('C',Feedback3/4,Target3/4,SerialFeedbackPort);
               SendTwoValues('c',PIDProcessDivider*16 + Disable1+(Disable2*2)+(Disable3*4),constrain(PWMout3,0,255),SerialFeedbackPort);
            }
            SerialFeedbackCounter = 0;
        }

        CommsTimeout++;
        if (CommsTimeout >= 60000)       // 15 second timeout ie 60000 * PROCESS_PERIOD_uS
        {
            CommsTimeout = 60000;        // So the counter doesn't overflow
            PowerScale = 9; 
        }
        else
        {
            PowerScale = 7;              // Number of bits to shift PID result (ie divide by 128 - approximate for /100)
        }
    }
}


//****************************************************************************************************************
//                                               TM1638 by Blame73                                              //
//****************************************************************************************************************  

void checkRangesAndButtons(){
#ifdef ENABLE_TM1638

    // Range 1 
    scaledCurrPosition1 = Feedback1;
    if (scaledCurrPosition1 != previousScaledCurrPosition1){
        maxPot1 = scaledCurrPosition1 > maxPot1 ? scaledCurrPosition1 : maxPot1;
        minPot1 = scaledCurrPosition1 < minPot1 ? scaledCurrPosition1 : minPot1;   
        limitDistancePot1 = scaledCurrPosition1 > 512 ? InputClipMax1 - scaledCurrPosition1 : scaledCurrPosition1 - InputClipMin1;
        maxLimitDistancePot1 = InputClipMax1 - scaledCurrPosition1 < maxLimitDistancePot1 ? InputClipMax1 - scaledCurrPosition1 : maxLimitDistancePot1;
        minLimitDistancePot1 = scaledCurrPosition1 - InputClipMin1 < minLimitDistancePot1 ? scaledCurrPosition1 - InputClipMin1 : minLimitDistancePot1;
    }

    //Range 2
    scaledCurrPosition2 = Feedback2;
    if (scaledCurrPosition2 != previousScaledCurrPosition2){
        maxPot2 = scaledCurrPosition2 > maxPot2 ? scaledCurrPosition2 : maxPot2;
        minPot2 = scaledCurrPosition2 < minPot2 ? scaledCurrPosition2 : minPot2;
        limitDistancePot2 = scaledCurrPosition2 > 512 ? InputClipMax2 - scaledCurrPosition2 : scaledCurrPosition2 - InputClipMin2;
        maxLimitDistancePot2 = InputClipMax2 - scaledCurrPosition2 < maxLimitDistancePot2 ? InputClipMax2 - scaledCurrPosition2 : maxLimitDistancePot2;
        minLimitDistancePot2 = scaledCurrPosition2 - InputClipMin2 < minLimitDistancePot2 ? scaledCurrPosition2 - InputClipMin2 : minLimitDistancePot2;                 
    }

    // Check buttons and send display updates
    currentMillis = millis();
    if (currentMillis - startMillis > btnPressedInterval){
        // DON'T check buttons every loop!!! It's very heavy, it lowers calculations from 4000/sec to 1000/sec! 
        buttons = moduleTM1638.getButtons();                
        if (buttons != 0) {
            switch (buttons){
               case button1:    // 1 > Current feedback values
                    TM1638_PAGE = 1;
                    showTM1638Message("Curr VAL", 1500);
                    break;                
               case button2:    // 2 >> Current distances from nearest clipping zones
                    showTM1638Message("CLIP dst", 1500);
                     TM1638_PAGE = 2;                            
                    break;
                case button3:    // 3 > Upper and lower clipping zones reached counters
                    TM1638_PAGE = 3;
                    showTM1638Message("CLIP cnt", 1500);
                    break;
                case button4:    // 4 > Min values reached
                    TM1638_PAGE = 4;
                    showTM1638Message("Min val", 1500);
                    break;
                case button5:    // 5 > Max values reached
                    TM1638_PAGE = 5;
                    showTM1638Message("MAx val", 1500);
                    break;
                case button6:    // 6 > Min distance from upper clipping zone reached (negative values when inside clipping zone)
                    TM1638_PAGE = 6;
                    showTM1638Message("EFFECtS", 1500);
                    break;  
                case button7:    // 7 > Enable/Disable                
                    #ifdef ENABLE_FASTLED
                        TM1638_PAGE = 7;
                        showTM1638Message(stripsDisabled ? "LEDS On" : "LEDS OFF", 1500);
                    #endif     
                    break;                            
                case button8:    // 8 > Change strips color mode
                    TM1638_PAGE = 8;
                    showTM1638Message("COL MODE", 500);
                    break;
                case button1_8:    // 9 > Reset values
                    TM1638_PAGE = 9;
                    showTM1638Message("rESEt", 1500);
                    break; 
                case button1_2:    // 21 (btn 2+1) > Store current value read from scaler as minimum scaler value  
                    TM1638_PAGE = 21;
                    showTM1638Message("SAVE SCL", 1500);
                    break;           
                case button2_3:    // 32 (btn 3+2) > Reset min scaler value to default
                    TM1638_PAGE = 32;
                    showTM1638Message("DFLT SCL", 1500);
                    break;
                case button6_7:    // 76 (btn 7+6) > Scaler values - % scaling and its value  
                    TM1638_PAGE = 76;
                    //showTM1638Message("SCALER", 500);
                    break;   
                case button8_9:    // 192 (btn 8+9) > Min distance from lower clipping zone reached (negative values when inside clipping zone)
                    TM1638_PAGE = 89;
                    showTM1638Message("dist dwn", 1500);
                    break;                                                                                                                 
                default:
                    showTM1638Message("not SEt");
                    break;                          
            }
            range_RA.addValue(50000);   // Add an high number to the average so that it will take some time
                                        // to turn the display back off again
            isDisplayOff = false;
            // #ifdef ENABLE_FASTLED
            //     if (!stripsDisabledbyUser)
            //         setStripsOnOff(isDisplayOff);
            //         Serial.println("1707");
            // #endif
          }
        if (!isDisplayOff)
            showRangeInfo();
        startMillis = currentMillis;
    }

    previousScaledCurrPosition1 = scaledCurrPosition1;
    previousScaledCurrPosition2 = scaledCurrPosition2;    
    
#endif 
}

void showRangeInfo(){     
#ifdef ENABLE_TM1638
    // for each rangePerLed we light up a led
    rangePerLed_WorkingZone = (InputClipMax1 -512) /5;                // four leds + all leds off when centered
    rangePerLed_ClippingZone = (CutoffLimitMax1 -InputClipMax1) /4;   // four leds
    
    // ***** Range1 *****
    // Inside Clipping Zones
    if ((scaledCurrPosition1 > InputClipMax1) || (scaledCurrPosition1 < InputClipMin1)){
        setTM1638LedsByPosition(scaledCurrPosition1, true, true); 

        // Cut Zones, motors disabled -- if messages enabled
        if (((scaledCurrPosition1 > CutoffLimitMax1) || (scaledCurrPosition1 < CutoffLimitMin1)) 
            && showAlertMessages) {
                showTM1638Message(errorMsg, 2000, false);         
        } 
        // Clipping Zones
        else {
            if (scaledCurrPosition1 > InputClipMax1){
                // I need 'stopCounting' to resume counting only when feedback goes outside the clipping zone
                // otherwise the counter would increase every cycle
                if (stopCountingMaxClipLimit1 == false){
                    maxClipLimitReachedPot1Count++;
                    stopCountingMaxClipLimit1 = true;
                }
            } 
            else if (scaledCurrPosition1 < InputClipMin1){
                if (stopCountingMinClipLimit1 == false){ 
                    minClipLimitReachedPot1Count++;
                    stopCountingMinClipLimit1 = true;
                }
            }   
            // Send a message if too close to CutOff Limits -- if messages enabled
            if (((CutoffLimitMax1 - scaledCurrPosition1 < maxCutoffWarningDistance) ||
                (scaledCurrPosition1 - CutoffLimitMin1 < minCutoffWarningDistance)) 
                && showAlertMessages) {
                    showTM1638Message(warningMsg, 1500, false);
            }
        }                
    } 
    // Correct working range
    else {
        setTM1638LedsByPosition(scaledCurrPosition1, false, true);
        stopCountingMaxClipLimit1 = false;
        stopCountingMinClipLimit1 = false;     
    }

    
    // ***** Range2 *****
    // Inside Clipping Zones
    if ((scaledCurrPosition2 > InputClipMax2) || (scaledCurrPosition2 < InputClipMin2)){
        setTM1638LedsByPosition(scaledCurrPosition2, true, false);
        
        // Cut Zones, motors disabled -- if messages enabled
        if (((scaledCurrPosition2 > CutoffLimitMax2) || (scaledCurrPosition2 < CutoffLimitMin2))
            && showAlertMessages) {
                showTM1638Message(errorMsg, 1500, false);
        } 
        // Clipping Zones
        else {
            if (scaledCurrPosition2 > InputClipMax2){
                if (stopCountingMaxClipLimit2 == false){
                    maxClipLimitReachedPot2Count++;
                    stopCountingMaxClipLimit2 = true;
                }
            } 
            else if (scaledCurrPosition2 < InputClipMin2){
                if (stopCountingMinClipLimit2 == false){ 
                    minClipLimitReachedPot2Count++;
                    stopCountingMinClipLimit2 = true;
                }
            }
            // Send a message if too close to CutOff Limits -- if messages enabled
            if (((CutoffLimitMax2 - scaledCurrPosition2 < maxCutoffWarningDistance) ||
                (scaledCurrPosition2 - CutoffLimitMin2 < minCutoffWarningDistance))
                && showAlertMessages) {
                    showTM1638Message(warningMsg, 1500, false);
            }
        }
    }  
    // Correct working range
    else {
        setTM1638LedsByPosition(scaledCurrPosition2, false, false);
        stopCountingMaxClipLimit2 = false;
        stopCountingMinClipLimit2 = false;        
    }

    // Display infos -- if a message's not being displayed or display hasn't been turned off
    if (!showingMessage && !isDisplayOff){
        // values rounded to 5 or 2 to have less flickering on the display
        // converted to string for easier grouping (two infos at a time for range1 and range2) and alignment
        // no performances loss in this case
        String  distance1, distance2, position1, position2, maxPosition1, maxPosition2, minPosition1, minPosition2, 
                maxDistance1, maxDistance2, minDistance1, minDistance2, scVal, valFromScaler;
        moduleTM1638.clearDisplay();

        switch (TM1638_PAGE) {
            case 1:    // 1 > Current feedback values           
                position1 = String(round((scaledCurrPosition1 -512) /5) * 5);
                position2 = String(round((scaledCurrPosition2 -512) /5) * 5);
                moduleTM1638.setDisplayToString(position1, 0, 4 -position1.length());
                moduleTM1638.setDisplayToString(position2, 0, 8 -position2.length());
                break;            
            case 2:    // 2 > Current distances from nearest clipping zones
                distance1 = String(round(limitDistancePot1 /5) * 5);
                distance2 = String(round(limitDistancePot2 /5) * 5);
                moduleTM1638.setDisplayToString(distance1, 0, 4 -distance1.length());
                moduleTM1638.setDisplayToString(distance2, 0, 8 -distance2.length());
                break;
            case 3:    // 2 > Upper and lower clipping zones reached counters
                moduleTM1638.setDisplayToString(String(maxClipLimitReachedPot1Count), 0, 2 -String(maxClipLimitReachedPot1Count).length());            
                moduleTM1638.setDisplayToString(String(minClipLimitReachedPot1Count), 0, 4 -String(minClipLimitReachedPot1Count).length());
                moduleTM1638.setDisplayToString(String(maxClipLimitReachedPot2Count), 0, 6 -String(maxClipLimitReachedPot2Count).length());                
                moduleTM1638.setDisplayToString(String(minClipLimitReachedPot2Count), 0, 8 -String(minClipLimitReachedPot2Count).length());
                break;
            case 4:    // 4 > Min values reached
                minPosition1 = String(round((minPot1 -512) /2) * 2);
                minPosition2 = String(round((minPot2 -512) /2) * 2);
                moduleTM1638.setDisplayToString(minPosition1, 0, 4 -minPosition1.length()); 
                moduleTM1638.setDisplayToString(minPosition2, 0, 8 -minPosition2.length());                           
                break;





            case 5:    // 5 > Max values reached
                maxPosition1 = String(round((maxPot1 -512) /2) * 2);
                maxPosition2 = String(round((maxPot2 -512) /2) * 2);
                moduleTM1638.setDisplayToString(maxPosition1, 0, 4 -maxPosition1.length());   
                moduleTM1638.setDisplayToString(maxPosition2, 0, 8 -maxPosition2.length());
                break;
            case 6:    // 6 > Min distance from upper clipping zone reached (negative values when inside clipping zone)
                #ifdef ENABLE_FASTLED
                setStripsOnOff(true, true, true);
                TM1638_PAGE = defaultPage;
                #endif
                // maxDistance1 = String(maxLimitDistancePot1);
                // maxDistance2 = String(maxLimitDistancePot2);
                // moduleTM1638.setDisplayToString(maxDistance1, 0, 4 -maxDistance1.length());
                // moduleTM1638.setDisplayToString(maxDistance2, 0, 8 -maxDistance2.length());
                break;
            case 7:    // 7 > Enable/Disable LED strips
                #ifdef ENABLE_FASTLED
                    setStripsOnOff(!stripsDisabled, true);
                    // Serial.println("1905");
                    TM1638_PAGE = defaultPage;
                #else
                     moduleTM1638.setDisplayToString("DISABLED");
                #endif
                break;
            case 8:    // 8 > Change strips color mode
                #ifdef ENABLE_FASTLED
                    sendValuesToSlave('P', 0); 
                    Serial.println("LEDs color mode changed");
                    TM1638_PAGE = defaultPage;
                #else
                     moduleTM1638.setDisplayToString("DISABLED");
                #endif




            case 9:    // 9 (btn 8+1) > Reset values
                resetRangeValues();            
                break;        
            case 21:    // 21 (btn 2+1) > Store current value read from scaler as minimum scaler value
                #ifdef ENABLE_SCALING            
                    scalerLowerLimit = scaler;               
                    writeIntIntoEEPROM(55, scalerLowerLimit);
                    delay(5); // some delay before reading the value from EEPROM
                    resetRangeValues(); 
                #else
                     moduleTM1638.setDisplayToString("DISABLED");
                #endif
                break;
            case 32:    // 32 (btn 3+2) > Reset min scaler value to default
                #ifdef ENABLE_SCALING
                    writeIntIntoEEPROM(55, defaultScalerLowerLimit);
                    delay(5); // some delay before reading the value from EEPROM
                    resetRangeValues(); 
                #else
                    moduleTM1638.setDisplayToString("DISABLED");
                #endif
                break;                
            case 76:    // 76 (btn 7+6) > Scaler values - % scaling and its value
                //moduleTM1638.setDisplayToString(String(scaledCurrPosition1));
                // Scaler values - % scaling and its value
                // self enabled when turning the pot
                #ifdef ENABLE_SCALING
                    scVal = String(100 - int(scaledValue *100)); 
                    valFromScaler = String(scaler); 
                    moduleTM1638.setDisplayToString(scVal, 0, 3 -scVal.length());
                    moduleTM1638.setDisplayToString(valFromScaler, 0, 8 -valFromScaler.length());                    
                #else
                     moduleTM1638.setDisplayToString("DISABLED");
                #endif
                break;
            case 89:    // 89 (btn 7+8) > Min distance from lower clipping zone reached (negative values when inside clipping zone)
                
                moduleTM1638.setDisplayToString(String(round(analogRead(PotScalerPin) /5) *5));
                
                //minDistance1 = String(minLimitDistancePot1);
                //minDistance2 = String(minLimitDistancePot2); 
                //moduleTM1638.setDisplayToString(minDistance1, 0, 4 -minDistance1.length());
                //moduleTM1638.setDisplayToString(minDistance2, 0, 8 -minDistance2.length());
                break;    
            default:
                // a different value shouldn't have been previously set
                break;
        }
    }
    else {
        currentMillisMessage = millis();
        if (currentMillisMessage - startMillisMessage > messageTimeout){
            showingMessage = false;
            startMillisMessage = currentMillisMessage;
        }      
    }

#endif
}

void showTM1638Message(String message){
#ifdef ENABLE_TM1638
    // overload, call function passing message, default timeout and default blinking
    showTM1638Message(message, 1000, false);
#endif
}

void showTM1638Message(String message, long timeout){
#ifdef ENABLE_TM1638
    // overload, call function passing message, timeout and default blinking
    showTM1638Message(message, timeout, false);
#endif
}

void showTM1638Message(String message, long timeout, bool blinking){
#ifdef ENABLE_TM1638

    messageTimeout = timeout;
    showingMessage = true;
    moduleTM1638.clearDisplay();
    if (!blinking){
        moduleTM1638.setDisplayToString(message);
    }
    // Has to be reimplemented not using delay()
    // So no blinking atm
    else {
        for (int i = 0; i <= timeout; i+= 400) {
            moduleTM1638.setDisplayToString(message);
            //delay(200);
            //moduleTM1638.clearDisplay();
            //delay(200);
        }
    }    
#endif 
}

// Color each led according to its position
void setTM1638LedsByPosition(int currentPosition, bool clippingZone, bool leftSide){
#ifdef ENABLE_TM1638

    byte led1 = leftSide ? 0 : 4;
    byte led2 = leftSide ? 1 : 5;
    byte led3 = leftSide ? 2 : 6;
    byte led4 = leftSide ? 3 : 7;
    int relativePosition;
    
    if (!clippingZone){      // normal working zone
       relativePosition = currentPosition > 512 ? currentPosition -512 : 512 -currentPosition;
       if (relativePosition < rangePerLed_WorkingZone){
                moduleTM1638.setLED(BLANK_LED, led1);
                moduleTM1638.setLED(BLANK_LED, led2);
                moduleTM1638.setLED(BLANK_LED, led3);
                moduleTM1638.setLED(BLANK_LED, led4);                     
        }
        else if (relativePosition < rangePerLed_WorkingZone *2){
                moduleTM1638.setLED(GREEN_LED, led1);
                moduleTM1638.setLED(BLANK_LED, led2);
                moduleTM1638.setLED(BLANK_LED, led3);
                moduleTM1638.setLED(BLANK_LED, led4);                     
        }
        else if(relativePosition < rangePerLed_WorkingZone *3){
                moduleTM1638.setLED(GREEN_LED, led1);
                moduleTM1638.setLED(GREEN_LED, led2);
                moduleTM1638.setLED(BLANK_LED, led3);
                moduleTM1638.setLED(BLANK_LED, led4);                     
        }
        else if(relativePosition < rangePerLed_WorkingZone *4){
                moduleTM1638.setLED(GREEN_LED, led1);
                moduleTM1638.setLED(GREEN_LED, led2);
                moduleTM1638.setLED(GREEN_LED, led3);
                moduleTM1638.setLED(BLANK_LED, led4);                     
        }
        else {      
                moduleTM1638.setLED(GREEN_LED, led1);
                moduleTM1638.setLED(GREEN_LED, led2);
                moduleTM1638.setLED(GREEN_LED, led3);
                moduleTM1638.setLED(ORANGE_LED, led4);
        }
    }
    else {                  // clipping zone
        relativePosition = currentPosition > 512 ? currentPosition -InputClipMax1 : InputClipMin1 -currentPosition;
        if (relativePosition < rangePerLed_ClippingZone){
                moduleTM1638.setLED(RED_LED, led1);
                moduleTM1638.setLED(BLANK_LED, led2);
                moduleTM1638.setLED(BLANK_LED, led3);
                moduleTM1638.setLED(BLANK_LED, led4);                     
        }
        else if(relativePosition < rangePerLed_ClippingZone *2){
                moduleTM1638.setLED(RED_LED, led1);
                moduleTM1638.setLED(RED_LED, led2);
                moduleTM1638.setLED(BLANK_LED, led3);
                moduleTM1638.setLED(BLANK_LED, led4);                     
        }
        else if(relativePosition < rangePerLed_ClippingZone *3){
                moduleTM1638.setLED(RED_LED, led1);
                moduleTM1638.setLED(RED_LED, led2);
                moduleTM1638.setLED(RED_LED, led3);
                moduleTM1638.setLED(BLANK_LED, led4);                     
        }
        else {      
                moduleTM1638.setLED(RED_LED, led1);
                moduleTM1638.setLED(RED_LED, led2);
                moduleTM1638.setLED(RED_LED, led3);
                moduleTM1638.setLED(RED_LED, led4);
        }
    }
#endif
}

void displayHandler(){
#ifdef ENABLE_TM1638
    range_RA.addValue(scaledCurrPosition1);
    float avg = range_RA.getFastAverage();
    // Check if current position is near average and near max e min values in the buffer to turn display off
    // (autoTurnOffDisplay == true has already been checked by the caller)
    if ((scaledCurrPosition1 < range_RA.getMinInBuffer() +15 && scaledCurrPosition1 > range_RA.getMaxInBuffer() -15) &&
        (abs(scaledCurrPosition1 - avg) < 15)){
        //turn display off
        if (!isDisplayOff){                     // No need to turn it off if it's already off
            moduleTM1638.clearDisplay();
            //Serial.println("2047");
            //moduleTM1638.setLEDs(BLANK_LED);            
            #ifdef ENABLE_FASTLED
                setStripsOnOff(true);
                // Serial.println("2049");                
            #endif     
            isDisplayOff = true;       
        }
    }    
    // if it's working and display is off, turn it back on
    else {  
        if (isDisplayOff){
            #ifdef ENABLE_FASTLED
                if (!stripsDisabledbyUser)
                    setStripsOnOff(false);
                    // Serial.println("2059");
            #endif
            isDisplayOff = false;
        }   
    }
#endif
}

void resetRangeValues(){
#ifdef ENABLE_TM1638
    maxPot1 = 0;
    minPot1= 1024;
    maxPot2 = 0;
    minPot2= 1024;
    maxClipLimitReachedPot1Count = 0;
    minClipLimitReachedPot1Count = 0;
    maxClipLimitReachedPot2Count = 0;
    minClipLimitReachedPot2Count = 0;
    limitDistancePot1 = 0;
    limitDistancePot2 = 0;
    maxLimitDistancePot1 = 1024;
    maxLimitDistancePot2 = 1024;
    minLimitDistancePot1 = 1024;
    minLimitDistancePot2 = 1024;
    TM1638_PAGE = defaultPage;
    readEepromTM1638();
    range_RA.addValue(50000);
#endif
}

void readEepromTM1638(){ 
#ifdef ENABLE_SCALING
    // read the lower scaler limit from the Arduino EEPROM
    if (!isnan(readIntFromEEPROM(55))){
        scalerLowerLimit = readIntFromEEPROM(55);
    } else {
        scalerLowerLimit = defaultScalerLowerLimit;
    }
#endif
}

int readIntFromEEPROM(int address){
  return (EEPROM.read(address) << 8) + EEPROM.read(address + 1);
}

void writeIntIntoEEPROM(int address, int number){ 
  int byte1 = number >> 8;
  int byte2 = number & 0xFF;
  EEPROM.write(address, byte1);
  EEPROM.write(address + 1, byte2);
}

//****************************************************************************************************************
//                                               FastLED by Blame73                                             //
//****************************************************************************************************************  

#ifdef ENABLE_FASTLED

// Start a transmission to the slave board checking any possible error
// A key/value pair is sent over
void sendValuesToSlave(const char key, int value){ 
    Wire.beginTransmission(I2CSlave);
    Wire.write(key);
    Wire.write(value);
    byte error = Wire.endTransmission(); 
    if (error){
        Serial.println("Couldn't send values to slave board");
        isSlaveAvailable = false;
    }
}

void setVariablesForArduinoSlave(){
    // Clipping Zones. Slave board has to know the clipping zones so to only use the valid zones
    // as range for the LED strips and to highlight when inside them
    sendValuesToSlave('G', InputClipMin1); delay(5); 
    sendValuesToSlave('H', InputClipMin2);   
    Serial.println("Parameters sent to slave board");
    // No other parameters are necessary atm
}

// Send a single byte transmission to the slave board to check its availablility
// Returns connection status
bool checkSlaveBoardStatus(){
    Wire.beginTransmission(I2CSlave);
    Wire.write(0);
    byte error = Wire.endTransmission();
    if (error){
        Serial.println("Slave board is unreachable");
        return false;
    }
    else{
        Serial.println("Slave board is alive!");
        setVariablesForArduinoSlave();
        return true;
    }        
}

// Event triggered by Arduino master board
void receiveEvent(int howMany)
{
    isSlaveAvailable = true;
    setVariablesForArduinoSlave();    
}

void reset(){ 
    asm volatile ("jmp 0");
}

void setStripsOnOff(bool disable){ setStripsOnOff(disable, false, false); }
void setStripsOnOff(bool disable, bool byUser){ setStripsOnOff(disable, false, false); }
void setStripsOnOff(bool disable, bool byUser, bool randomEffects){
    if (randomEffects){
        sendValuesToSlave('O', 2);
        stripsDisabledbyUser = true;
    }
    else{
        if (disable){
            sendValuesToSlave('O', 0);            
        }
        else{
            sendValuesToSlave('O', 1);
          
        }
        //Serial.println(disable ? "Turned strip OFF" : "Turned strip ON");  // can't use it because it turns my TM1638 display off! (...)
        stripsDisabledbyUser = byUser && disable;       //if the user disabled the strips, only the user can re-enable them
                                                        //it can't be automatically done by timeout end          
        stripsDisabled = disable;        
    }
}

#endif